import pytest
from numpy.random import RandomState

from darts.tests.conftest import TORCH_AVAILABLE

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )
import torch

from darts.utils.torch import random_method


# use a simple torch model mock
class TorchModelMock:
    @random_method
    def __init__(self, some_params=None, **kwargs):
        self.model = torch.randn(5)
        # super().__init__()

    @random_method
    def fit(self, some_params=None):
        self.fit_value = torch.randn(5)


class TestRandomMethod:
    def test_it_raises_error_if_used_on_function(self):
        with pytest.raises(ValueError):

            @random_method
            def a_random_function():
                pass

    def test_model_is_random_by_default(self):
        model1 = TorchModelMock()
        model2 = TorchModelMock()
        assert not torch.equal(model1.model, model2.model)

    def test_model_is_random_when_None_random_state_specified(self):
        model1 = TorchModelMock(random_state=None)
        model2 = TorchModelMock(random_state=None)
        assert not torch.equal(model1.model, model2.model)

    def helper_test_reproducibility(self, model1, model2):
        assert torch.equal(model1.model, model2.model)

        model1.fit()
        model2.fit()
        assert torch.equal(model1.fit_value, model2.fit_value)

    def test_model_is_reproducible_when_seed_specified(self):
        model1 = TorchModelMock(random_state=42)
        model2 = TorchModelMock(random_state=42)
        self.helper_test_reproducibility(model1, model2)

    def test_model_is_reproducible_when_random_instance_specified(self):
        model1 = TorchModelMock(random_state=RandomState(42))
        model2 = TorchModelMock(random_state=RandomState(42))
        self.helper_test_reproducibility(model1, model2)

    def test_model_is_different_for_different_seeds(self):
        model1 = TorchModelMock(random_state=42)
        model2 = TorchModelMock(random_state=43)
        assert not torch.equal(model1.model, model2.model)

    def test_model_is_different_for_different_random_instance(self):
        model1 = TorchModelMock(random_state=RandomState(42))
        model2 = TorchModelMock(random_state=RandomState(43))
        assert not torch.equal(model1.model, model2.model)

    def helper_test_successive_call_are_different(self, model):
        # different between init and fit
        model.fit()
        assert not torch.equal(model.model, model.fit_value)

        # different between 2 fit
        old_fit_value = model.fit_value.clone()
        model.fit()
        assert not torch.equal(model.fit_value, old_fit_value)

    def test_successive_call_to_rng_are_different_when_seed_specified(self):
        model = TorchModelMock(random_state=42)
        self.helper_test_successive_call_are_different(model)

    def test_successive_call_to_rng_are_different_when_random_instance_specified(
        self,
    ):
        model = TorchModelMock(random_state=RandomState(42))
        self.helper_test_successive_call_are_different(model)

    def test_no_side_effect_between_rng_with_seeds(self):
        model = TorchModelMock(random_state=42)
        model.fit()
        fit_value = model.fit_value.clone()

        model = TorchModelMock(random_state=42)
        model2 = TorchModelMock(random_state=42)
        model2.fit()
        model.fit()

        assert torch.equal(model.fit_value, fit_value)

    def test_no_side_effect_between_rng_with_random_instance(self):
        model = TorchModelMock(random_state=RandomState(42))
        model.fit()
        fit_value = model.fit_value.clone()

        model = TorchModelMock(random_state=RandomState(42))
        model2 = TorchModelMock(random_state=RandomState(42))
        model2.fit()
        model.fit()

        assert torch.equal(model.fit_value, fit_value)



"""
Kalman Filter
-------------
"""

from abc import ABC
from copy import deepcopy
from typing import Optional

import numpy as np
import pandas as pd
from nfoursid.kalman import Kalman
from nfoursid.nfoursid import NFourSID

from darts.logging import raise_if, raise_if_not
from darts.models.filtering.filtering_model import FilteringModel
from darts.timeseries import TimeSeries


class KalmanFilter(FilteringModel, ABC):
    def __init__(self, dim_x: int = 1, kf: Optional[Kalman] = None):
        """
        This model implements a Kalman filter over a time series.

        The key method is `KalmanFilter.filter()`.
        It considers the provided time series as containing (possibly noisy) observations z obtained from a
        (possibly noisy) linear dynamical system with hidden state x. The function `filter(series)` returns a new
        `TimeSeries` describing the distribution of the output z (without noise), as inferred by the Kalman filter from
        sequentially observing z from `series`, and the dynamics of the linear system of order dim_x.

        This filter also supports missing values in the observation series, in which case the underlying
        Kalman filter will carry on using its mean estimate.

        The method `KalmanFilter.fit()` is used to initialize the Kalman filter by estimating the state space model of
        a linear dynamical system and the covariance matrices of the process and measurement noise using the N4SID
        algorithm.

        This implementation uses Kalman from the NFourSID package. More information can be found here:
        https://nfoursid.readthedocs.io/en/latest/source/kalman.html.

        The dimensionality of the measurements z and optional control signal (covariates) u is automatically inferred
        upon calling `filter()`.

        Parameters
        ----------
        dim_x : int
            Size of the Kalman filter state vector.
        kf : nfoursid.kalman.Kalman
            Optionally, an instance of `nfoursid.kalman.Kalman`.
            If this is provided, the parameter dim_x is ignored. This instance will be copied for every
            call to `filter()`, so the state is not carried over from one time series to another across several
            calls to `filter()`.
            The dimensionalities of the filter must match those of the `TimeSeries` used when calling `filter()`.
        """
        # TODO: Add support for x_init. Needs reimplementation of NFourSID.

        super().__init__()

        if kf is None:
            self.kf = None
            self.dim_x = dim_x
            self._kf_provided = False
        else:
            self.kf = kf
            self.dim_u = kf.state_space.u_dim
            self.dim_x = kf.state_space.x_dim
            self.dim_y = kf.state_space.y_dim
            self._kf_provided = True
            if self.dim_u > 0:
                self._expect_covariates = True

    def __str__(self):
        return f"KalmanFilter(dim_x={self.dim_x})"

    def fit(
        self,
        series: TimeSeries,
        covariates: Optional[TimeSeries] = None,
        num_block_rows: Optional[int] = None,
    ) -> "KalmanFilter":
        """
        Initializes the Kalman filter using the N4SID algorithm.

        Parameters
        ----------
        series : TimeSeries
            The series of outputs (observations) used to infer the underlying state space model.
            This must be a deterministic series (containing one sample).
        covariates : Optional[TimeSeries]
            An optional series of inputs (control signal) that will also be used to infer the underlying state space
            model. This must be a deterministic series (containing one sample).
        num_block_rows : Optional[int]
            The number of block rows to use in the block Hankel matrices used in the N4SID algorithm.
            See the documentation of nfoursid.nfoursid.NFourSID for more information.
            If not provided, the dimensionality of the state space model will be used, with a maximum of 10.

        Returns
        -------
        self
            Fitted Kalman filter.
        """
        if covariates is not None:
            self._expect_covariates = True
            covariates = covariates.slice_intersect(series)
            raise_if_not(
                series.has_same_time_as(covariates),
                "The number of timesteps in the series and the covariates must match.",
            )

        # TODO: Handle multiple timeseries. Needs reimplementation of NFourSID?
        self.dim_y = series.width
        outputs = series.to_dataframe(copy=False)
        outputs.columns = [f"y_{i}" for i in outputs.columns]

        if covariates is not None:
            self.dim_u = covariates.width
            inputs = covariates.to_dataframe(copy=False)
            inputs.columns = [f"u_{i}" for i in inputs.columns]
            input_columns = list(inputs.columns)
            measurements = pd.concat([outputs, inputs], axis=1)
        else:
            measurements = outputs
            input_columns = None

        if num_block_rows is None:
            num_block_rows = max(10, self.dim_x)
        nfoursid = NFourSID(
            measurements,
            output_columns=list(outputs.columns),
            input_columns=input_columns,
            num_block_rows=num_block_rows,
        )
        nfoursid.subspace_identification()
        state_space_identified, covariance_matrix = nfoursid.system_identification(
            rank=self.dim_x
        )

        self.kf = Kalman(state_space_identified, covariance_matrix)

        return self

    def filter(
        self,
        series: TimeSeries,
        covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
    ) -> TimeSeries:
        """
        Sequentially applies the Kalman filter on the provided series of observations.

        Parameters
        ----------
        series : TimeSeries
            The series of outputs (observations) used to infer the underlying outputs according to the specified Kalman
            process. This must be a deterministic series (containing one sample).
        covariates : Optional[TimeSeries]
            An optional series of inputs (control signal), necessary if the Kalman filter was initialized with
            covariates. This must be a deterministic series (containing one sample).
        num_samples : int, default: 1
            The number of samples to generate from the inferred distribution of the output z. If this is set to 1, the
            output is a `TimeSeries` containing a single sample using the mean of the distribution.

        Returns
        -------
        TimeSeries
            A (stochastic) `TimeSeries` of the inferred output z, of the same width as the input series.
        """
        super().filter(series)

        raise_if(
            self.kf is None,
            "The Kalman filter has not been fitted yet. Call `fit()` first "
            "or provide Kalman filter in constructor.",
        )

        raise_if_not(
            series.width == self.dim_y,
            "The provided TimeSeries dimensionality does not match "
            "the output dimensionality of the Kalman filter.",
        )

        raise_if(
            covariates is not None and not self._expect_covariates,
            "Covariates were provided, but the Kalman filter was not fitted with covariates.",
        )

        if self._expect_covariates:
            raise_if(
                covariates is None,
                "The Kalman filter was fitted with covariates, but these were not provided.",
            )

            raise_if_not(
                covariates.is_deterministic,
                "The covariates must be deterministic (observations).",
            )

            covariates = covariates.slice_intersect(series)
            raise_if_not(
                series.has_same_time_as(covariates),
                "The number of timesteps in the series and the covariates must match.",
            )

        kf = deepcopy(self.kf)

        y_values = series.values(copy=False)
        if self._expect_covariates:
            u_values = covariates.values(copy=False)

            # set control signal to 0 if it contains NaNs:
            u_values = np.nan_to_num(u_values, copy=True, nan=0.0)
        else:
            u_values = np.zeros((len(y_values), 0))

        # For each time step, we'll sample "n_samples" from a multivariate Gaussian
        # whose mean vector and covariance matrix come from the Kalman filter.
        sampled_outputs = np.zeros((len(y_values), self.dim_y, num_samples))

        for i in range(len(y_values)):
            y = y_values[i, :].reshape(-1, 1)
            u = u_values[i, :].reshape(-1, 1)

            if np.isnan(y).any():
                y = None

            kf.step(y, u)
            mean_vec = kf.y_filtereds[-1].reshape(
                self.dim_y,
            )

            if num_samples == 1:
                sampled_outputs[i, :, 0] = mean_vec
            else:
                # The measurement covariance matrix is given by the sum of the covariance matrix of the
                # state estimate (transformed by C) and the covariance matrix of the measurement noise.
                cov_matrix = (
                    kf.state_space.c @ kf.p_filtereds[-1] @ kf.state_space.c.T + kf.r
                )
                sampled_outputs[i, :, :] = np.random.multivariate_normal(
                    mean_vec, cov_matrix, size=num_samples
                ).T

        return series.with_values(sampled_outputs)



"""
TFT Explainer for Temporal Fusion Transformer (TFTModel)
------------------------------------

The `TFTExplainer` uses a trained :class:`TFTModel <darts.models.forecasting.tft_model.TFTModel>` and extracts the
explainability information from the model.

- :func:`plot_variable_selection() <TFTExplainer.plot_variable_selection>` plots the variable selection weights for
  each of the input features.
  - encoder importance: historic part of target, past covariates and historic part of future covariates
  - decoder importance: future part of future covariates
  - static covariates importance: the numeric and catageorical static covariates importance

- :func:`plot_attention() <TFTExplainer.plot_attention>` plots the transformer attention that the `TFTModel` applies
  on the given past and future input. The attention is aggregated over all attention heads.

The attention and feature importance values can be extracted using the :class:`TFTExplainabilityResult
<darts.explainability.explainability_result.TFTExplainabilityResult>` returned by
:func:`explain() <TFTExplainer.explain>`. An example of this is shown in the method description.

We also show how to use the `TFTExplainer` in the example notebook of the `TFTModel` `here
<https://unit8co.github.io/darts/examples/13-TFT-examples.html#Explainability>`_.
"""

from collections.abc import Sequence
from typing import Literal, Optional, Union

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from torch import Tensor

from darts import TimeSeries
from darts.explainability import TFTExplainabilityResult
from darts.explainability.explainability import _ForecastingModelExplainer
from darts.logging import get_logger, raise_log
from darts.models import TFTModel
from darts.utils.utils import generate_index

logger = get_logger(__name__)


class TFTExplainer(_ForecastingModelExplainer):
    model: TFTModel

    def __init__(
        self,
        model: TFTModel,
        background_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        background_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        background_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
    ):
        """
        Explainer class for the `TFTModel`.

        **Definitions**

        - A background series is a `TimeSeries` that is used as a default for generating the explainability result
          (if no `foreground` is passed to :func:`explain() <TFTExplainer.explain>`).
        - A foreground series is a `TimeSeries` that can be passed to :func:`explain() <TFTExplainer.explain>` to use
          instead of the background for generating the explainability result.

        Parameters
        ----------
        model
            The fitted `TFTModel` to be explained.
        background_series
            Optionally, a series or list of series to use as a default target series for the explanations.
            Optional if `model` was trained on a single target series. By default, it is the `series` used at fitting
            time.
            Mandatory if `model` was trained on multiple (sequence of) target series.
        background_past_covariates
            Optionally, a past covariates series or list of series to use as a default past covariates series
            for the explanations. The same requirements apply as for `background_series` .
        background_future_covariates
            Optionally, a future covariates series or list of series to use as a default future covariates series
            for the explanations. The same requirements apply as for `background_series`.

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.explainability.tft_explainer import TFTExplainer
        >>> from darts.models import TFTModel
        >>> series = AirPassengersDataset().load()
        >>> model = TFTModel(
        >>>     input_chunk_length=12,
        >>>     output_chunk_length=6,
        >>>     add_encoders={"cyclic": {"future": ["hour"]}}
        >>> )
        >>> model.fit(series)
        >>> # create the explainer and generate explanations
        >>> explainer = TFTExplainer(model)
        >>> results = explainer.explain()
        >>> # plot the results
        >>> explainer.plot_attention(results, plot_type="all")
        >>> explainer.plot_variable_selection(results)
        """
        super().__init__(
            model,
            background_series=background_series,
            background_past_covariates=background_past_covariates,
            background_future_covariates=background_future_covariates,
            requires_background=True,
            requires_covariates_encoding=False,
            check_component_names=False,
            test_stationarity=False,
        )
        # add the relative index that generated inside the model (not in the input data)
        if model.add_relative_index:
            if self.future_covariates_components is not None:
                self.future_covariates_components.append("add_relative_index")
            else:
                self.future_covariates_components = ["add_relative_index"]

    def explain(
        self,
        foreground_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        foreground_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        foreground_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        horizons: Optional[Sequence[int]] = None,
        target_components: Optional[Sequence[str]] = None,
    ) -> TFTExplainabilityResult:
        """Returns the :class:`TFTExplainabilityResult
        <darts.explainability.explainability_result.TFTExplainabilityResult>` result for all series in
        `foreground_series`. If `foreground_series` is `None`, will use the `background` input
        from `TFTExplainer` creation (either the `background` passed to creation, or the series stored in the
        `TFTModel` in case it was only trained on a single series).
        For each series, the results contain the attention heads, encoder variable importances, decoder variable
        importances, and static covariates importances.

        Parameters
        ----------
        foreground_series
            Optionally, one or a sequence of target `TimeSeries` to be explained. Can be multivariate.
            If not provided, the background `TimeSeries` will be explained instead.
        foreground_past_covariates
            Optionally, one or a sequence of past covariates `TimeSeries` if required by the forecasting model.
        foreground_future_covariates
            Optionally, one or a sequence of future covariates `TimeSeries` if required by the forecasting model.
        horizons
            This parameter is not used by the `TFTExplainer`.
        target_components
            This parameter is not used by the `TFTExplainer`.

        Returns
        -------
        TFTExplainabilityResult
            The explainability result containing the attention heads, encoder variable importances, decoder variable
            importances, and static covariates importances.

        Examples
        --------
        >>> explainer = TFTExplainer(model)  # requires `background` if model was trained on multiple series

        Optionally, give a foreground input to generate the explanation on a new input.
        Otherwise, leave it empty to compute the explanation on the background from `TFTExplainer` creation

        >>> explain_results = explainer.explain(
        >>>     foreground_series=foreground_series,
        >>>     foreground_past_covariates=foreground_past_covariates,
        >>>     foreground_future_covariates=foreground_future_covariates,
        >>> )
        >>> attn = explain_results.get_attention()
        >>> importances = explain_results.get_feature_importances()
        """
        if target_components is not None or horizons is not None:
            logger.warning(
                "`horizons`, and `target_components` are not supported by `TFTExplainer` and will be ignored."
            )
        super().explain(
            foreground_series, foreground_past_covariates, foreground_future_covariates
        )
        (
            foreground_series,
            foreground_past_covariates,
            foreground_future_covariates,
            _,
            _,
            _,
            _,
        ) = self._process_foreground(
            foreground_series,
            foreground_past_covariates,
            foreground_future_covariates,
        )
        horizons, _ = self._process_horizons_and_targets(None, None)
        preds = self.model.predict(
            n=self.n,
            series=foreground_series,
            past_covariates=foreground_past_covariates,
            future_covariates=foreground_future_covariates,
        )
        # get the weights and the attention head from the trained model for the prediction
        # aggregate over attention heads
        attention_heads = (
            self.model.model._attn_out_weights.detach().cpu().numpy().sum(axis=-2)
        )
        # get the variable importances (pd.DataFrame with rows corresponding to the number of input series)
        encoder_importance = self._encoder_importance
        decoder_importance = self._decoder_importance
        static_covariates_importance = self._static_covariates_importance

        horizon_idx = [h - 1 for h in horizons]

        results = []
        icl = self.model.input_chunk_length
        for idx, (series, pred_series) in enumerate(zip(foreground_series, preds)):
            times = series.time_index[-icl:].union(pred_series.time_index)
            attention = TimeSeries.from_times_and_values(
                values=np.take(attention_heads[idx], horizon_idx, axis=0).T,
                times=times,
                columns=[f"horizon {str(i)}" for i in horizons],
            )
            results.append({
                "attention": attention,
                "encoder_importance": encoder_importance.iloc[idx : idx + 1],
                "decoder_importance": decoder_importance.iloc[idx : idx + 1],
                "static_covariates_importance": static_covariates_importance.iloc[
                    idx : idx + 1
                ],
            })
        return TFTExplainabilityResult(
            explanations=results[0] if len(results) == 1 else results
        )

    def plot_variable_selection(
        self,
        expl_result: TFTExplainabilityResult,
        fig_size=None,
        max_nr_series: int = 5,
    ):
        """Plots the variable selection / feature importances of the `TFTModel` based on the input.
        The figure includes three subplots:

        - encoder importances: contains the past target, past covariates, and historic future covariates importance
          on the encoder (input chunk)
        - decoder importances: contains the future covariates importance on the decoder (output chunk)
        - static covariates importances: contains the numeric and / or categorical static covariates importance

        Parameters
        ----------
        expl_result
            A `TFTExplainabilityResult` object. Corresponds to the output of :func:`explain() <TFTExplainer.explain>`.
        fig_size
            The size of the figure to be plotted.
        max_nr_series
            The maximum number of plots to show in case `expl_result` was computed on multiple series.
        """
        encoder_importance = expl_result.get_encoder_importance()
        decoder_importance = expl_result.get_decoder_importance()
        static_covariates_importance = expl_result.get_static_covariates_importance()
        if not isinstance(encoder_importance, list):
            encoder_importance = [encoder_importance]
            decoder_importance = [decoder_importance]
            static_covariates_importance = [static_covariates_importance]

        uses_static_covariates = not static_covariates_importance[0].empty
        for idx, (enc_imp, dec_imp, stc_imp) in enumerate(
            zip(encoder_importance, decoder_importance, static_covariates_importance)
        ):
            # plot the encoder and decoder weights
            fig, axes = plt.subplots(
                nrows=3 if uses_static_covariates else 2, sharex=True, figsize=fig_size
            )
            self._plot_cov_selection(
                enc_imp, title="Encoder variable importance", ax=axes[0]
            )
            axes[0].set_xlabel("")
            self._plot_cov_selection(
                dec_imp, title="Decoder variable importance", ax=axes[1]
            )
            if uses_static_covariates:
                axes[1].set_xlabel("")
                self._plot_cov_selection(
                    stc_imp,
                    title="Static variable importance",
                    ax=axes[2],
                )
            fig.tight_layout()
            plt.show()

            if idx + 1 == max_nr_series:
                break

    def plot_attention(
        self,
        expl_result: TFTExplainabilityResult,
        plot_type: Optional[Literal["all", "time", "heatmap"]] = "all",
        show_index_as: Literal["relative", "time"] = "relative",
        ax: Optional[matplotlib.axes.Axes] = None,
        max_nr_series: int = 5,
        show_plot: bool = True,
    ) -> matplotlib.axes.Axes:
        """Plots the attention heads of the `TFTModel`.

        Parameters
        ----------
        expl_result
            A `TFTExplainabilityResult` object. Corresponds to the output of :func:`explain() <TFTExplainer.explain>`.
        plot_type
            The type of attention head plot. One of ("all", "time", "heatmap").
            If "all", will plot the attention per horizon (given the horizons in the `TFTExplainabilityResult`).
            The maximum horizon corresponds to the `output_chunk_length` of the trained `TFTModel`.
            If "time", will plot the mean attention over all horizons.
            If "heatmap", will plot the attention per horizon on a heat map. The horizons are shown on the y-axis,
            and times / relative indices on the x-axis.
        show_index_as
            The type of index to be shown. One of ("relative", "time").
            If "relative", will plot the x-axis from `(-input_chunk_length, output_chunk_length - 1)`. `0` corresponds
            to the first prediction point.
            If "time", will plot the x-axis with the actual time index (or range index) of the corresponding
            `TFTExplainabilityResult`.
        ax
            Optionally, an axis to plot on. Only effective on a single `expl_result`.
        max_nr_series
            The maximum number of plots to show in case `expl_result` was computed on multiple series.
        show_plot
            Whether to show the plot.
        """
        single_series = False
        attentions = expl_result.get_explanation(component="attention")
        if isinstance(attentions, TimeSeries):
            attentions = [attentions]
            single_series = True

        for idx, attention in enumerate(attentions):
            if ax is None or not single_series:
                fig, ax = plt.subplots()
            if show_index_as == "relative":
                x_ticks = generate_index(
                    start=-self.model.input_chunk_length, end=self.n - 1
                )
                attention = TimeSeries.from_times_and_values(
                    times=generate_index(
                        start=-self.model.input_chunk_length, end=self.n - 1
                    ),
                    values=attention.values(copy=False),
                    columns=attention.components,
                )
                x_label = "Index relative to first prediction point"
            elif show_index_as == "time":
                x_ticks = attention.time_index
                x_label = "Time index"
            else:
                x_label, x_ticks = None, None
                raise_log(
                    ValueError("`show_index_as` must either be 'relative', or 'time'.")
                )

            prediction_start_color = "red"
            if plot_type == "all":
                ax_title = "Attention per Horizon"
                y_label = "Attention"
                attention.plot(max_nr_components=-1, ax=ax)
            elif plot_type == "time":
                ax_title = "Mean Attention"
                y_label = "Attention"
                attention.mean(1).plot(label="Mean Attention Head", ax=ax)
            elif plot_type == "heatmap":
                ax_title = "Attention Heat Map"
                y_label = "Horizon"

                # generate a heat map
                x, y = np.meshgrid(x_ticks, np.arange(1, self.n + 1))
                c = ax.pcolormesh(x, y, attention.values().transpose(), cmap="hot")
                ax.axis([x.min(), x.max(), y.max(), y.min()])
                prediction_start_color = "lightblue"
                fig.colorbar(c, ax=ax, orientation="horizontal")
            else:
                raise raise_log(
                    ValueError("`plot_type` must be either 'all', 'time' or 'heatmap'"),
                    logger=logger,
                )

            # draw the prediction start point
            y_min, y_max = ax.get_ylim()
            ax.vlines(
                x=x_ticks[-self.n],
                ymin=y_min,
                ymax=y_max,
                label="prediction start",
                ls="dashed",
                lw=2,
                colors=prediction_start_color,
            )
            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)
            title_suffix = "" if single_series else f": series index {idx}"
            ax.set_title(ax_title + title_suffix)
            ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
            if show_plot:
                plt.show()

            if idx + 1 == max_nr_series:
                break
        return ax

    @property
    def _encoder_importance(self) -> pd.DataFrame:
        """Returns the encoder variable importance of the TFT model.

        The encoder_weights are calculated for the past inputs of the model.
        The encoder_importance contains the weights of the encoder variable selection network.
        The encoder variable selection network is used to select the most important static and time dependent
        covariates. It provides insights which variable are most significant for the prediction problem.
        See section 4.2 of the paper for more details.

        Returns
        -------
        pd.DataFrame
            The encoder variable importance.
        """
        return self._get_importance(
            weight=self.model.model._encoder_sparse_weights,
            names=self.model.model.encoder_variables,
        )

    @property
    def _decoder_importance(self) -> pd.DataFrame:
        """Returns the decoder variable importance of the TFT model.

        The decoder_weights are calculated for the known future inputs of the model.
        The decoder_importance contains the weights of the decoder variable selection network.
        The decoder variable selection network is used to select the most important static and time dependent
        covariates. It provides insights which variable are most significant for the prediction problem.
        See section 4.2 of the paper for more details.

        Returns
        -------
        pd.DataFrame
            The importance of the decoder variables.
        """
        return self._get_importance(
            weight=self.model.model._decoder_sparse_weights,
            names=self.model.model.decoder_variables,
        )

    @property
    def _static_covariates_importance(self) -> pd.DataFrame:
        """Returns the static covariates importance of the TFT model.

        The static covariate importances are calculated for the static inputs of the model (numeric and / or
        categorical). The static variable selection network is used to select the most important static covariates.
        It provides insights which variable are most significant for the prediction problem.
        See section 4.2, and 4.3 of the paper for more details.

        Returns
        -------
        pd.DataFrame
            The static covariates importance.
        """
        return self._get_importance(
            weight=self.model.model._static_covariate_var,
            names=self.model.model.static_variables,
        )

    def _get_importance(
        self,
        weight: Tensor,
        names: list[str],
        n_decimals=3,
    ) -> pd.DataFrame:
        """Returns the encoder or decoder variable of the TFT model.

        Parameters
        ----------
        weights
            The weights of the encoder or decoder of the trained TFT model.
        names
            The encoder or decoder names saved in the TFT model class.
        n_decimals
            The number of decimals to round the importance to.

        Returns
        -------
        pd.DataFrame
            The importance of the variables.
        """
        if weight is None:
            return pd.DataFrame()

        # static covariates aggregation needs expansion in first dimension
        if weight.ndim == 3:
            weight = weight.unsqueeze(1)

        # transform the encoder/decoder weights to percentages, rounded to n_decimals
        weights_percentage = (
            weight.detach().cpu().numpy().mean(axis=1).squeeze(axis=1).round(n_decimals)
            * 100
        )
        # create a dataframe with the variable names and the weights
        name_mapping = self._name_mapping
        importance = pd.DataFrame(
            weights_percentage,
            columns=[name_mapping[name] for name in names],
        )

        # return the importance sorted descending
        return importance.transpose().sort_values(0, ascending=True).transpose()

    @property
    def _name_mapping(self) -> dict[str, str]:
        """Returns the feature name mapping of the TFT model.

        Returns
        -------
        Dict[str, str]
            The feature name mapping. For example
            {
                'target_0': 'ice cream',
                'past_covariate_0': 'heater',
                'past_covariate_1': 'year',
                'past_covariate_2': 'month',
                'future_covariate_0': 'darts_enc_fc_cyc_month_sin',
                'future_covariate_1': 'darts_enc_fc_cyc_month_cos',
             }
        """

        def map_cols(comps, name, suffix):
            comps = comps if comps is not None else []
            return {
                f"{name}_{i}": colname + f"_{suffix}" for i, colname in enumerate(comps)
            }

        return {
            **map_cols(self.target_components, "target", "target"),
            **map_cols(
                self.static_covariates_components, "static_covariate", "statcov"
            ),
            **map_cols(self.past_covariates_components, "past_covariate", "pastcov"),
            **map_cols(self.future_covariates_components, "future_covariate", "futcov"),
        }

    @staticmethod
    def _plot_cov_selection(
        importance: pd.DataFrame,
        title: str = "Variable importance",
        ax: Optional[matplotlib.axes.Axes] = None,
    ):
        """Plots the variable importance of the TFT model.

        Parameters
        ----------
        importance
            The encoder / decoder importance.
        title
            The title of the plot.
        ax
            Optionally, an axis to plot on. Otherwise, will create and plot on a new axis.
        """
        if ax is None:
            _, ax = plt.subplots()
        ax.barh(importance.columns.tolist(), importance.values[0].tolist())
        ax.set_title(title)
        ax.set_ylabel("Variable", fontsize=12)
        ax.set_xlabel("Variable importance in %")
        return ax



"""
darts
-----
"""

import os

import matplotlib as mpl
from matplotlib import cycler

from darts.timeseries import TimeSeries, concatenate, slice_intersect

__version__ = "0.35.0"

colors = cycler(
    color=["black", "003DFD", "b512b8", "11a9ba", "0d780f", "f77f07", "ba0f0f"]
)

u8plots_mplstyle = {
    "font.family": "sans serif",
    "axes.edgecolor": "black",
    "axes.grid": True,
    "axes.labelcolor": "#333333",
    "axes.labelweight": 600,
    "axes.linewidth": 1,
    "axes.prop_cycle": colors,
    "axes.spines.top": False,
    "axes.spines.right": False,
    "axes.spines.bottom": False,
    "axes.spines.left": False,
    "grid.color": "#dedede",
    "legend.frameon": False,
    "lines.linewidth": 1.3,
    "xtick.color": "#333333",
    "xtick.labelsize": "small",
    "ytick.color": "#333333",
    "ytick.labelsize": "small",
    "xtick.bottom": False,
}


if os.getenv("DARTS_CONFIGURE_MATPLOTLIB", "1") != "0":
    mpl.rcParams.update(u8plots_mplstyle)

__all__ = ["TimeSeries", "concatenate", "slice_intersect"]



import os
from itertools import product

import numpy as np
import pytest
from sklearn.preprocessing import MaxAbsScaler

from darts.dataprocessing.transformers import Scaler
from darts.datasets import AirPassengersDataset
from darts.metrics import smape
from darts.models import LinearRegressionModel
from darts.tests.conftest import OPTUNA_AVAILABLE, TORCH_AVAILABLE, tfm_kwargs

if not OPTUNA_AVAILABLE:
    pytest.skip(
        f"Optuna not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )

import optuna

if TORCH_AVAILABLE:
    import torch
    from pytorch_lightning.callbacks import Callback, EarlyStopping

    # hacky workaround found in https://github.com/Lightning-AI/pytorch-lightning/issues/17485
    # to avoid import of both lightning and pytorch_lightning
    class PatchedPruningCallback(
        optuna.integration.PyTorchLightningPruningCallback, Callback
    ):
        pass

    from darts.models import TCNModel
    from darts.utils.likelihood_models.torch import GaussianLikelihood


class TestOptuna:
    series = AirPassengersDataset().load().astype(np.float32)

    val_length = 36
    train, val = series.split_after(val_length)

    # scale
    scaler = Scaler(MaxAbsScaler())
    train = scaler.fit_transform(train)
    val = scaler.transform(val)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_optuna_torch_model(self, tmpdir_fn):
        """Check that optuna works as expected with a torch-based model"""

        # define objective function
        def objective(trial):
            # select input and output chunk lengths
            in_len = trial.suggest_int("in_len", 4, 8)
            out_len = trial.suggest_int("out_len", 1, 3)

            # Other hyperparameters
            kernel_size = trial.suggest_int("kernel_size", 2, 3)
            num_filters = trial.suggest_int("num_filters", 1, 2)
            lr = trial.suggest_float("lr", 5e-5, 1e-3, log=True)
            include_year = trial.suggest_categorical("year", [False, True])

            # throughout training we'll monitor the validation loss for both pruning and early stopping
            pruner = PatchedPruningCallback(trial, monitor="val_loss")
            early_stopper = EarlyStopping(
                "val_loss", min_delta=0.001, patience=3, verbose=True
            )

            # optionally also add the (scaled) year value as a past covariate
            if include_year:
                encoders = {
                    "datetime_attribute": {"past": ["year"]},
                    "transformer": Scaler(),
                }
            else:
                encoders = None

            # reproducibility
            torch.manual_seed(42)

            # build the TCN model
            model = TCNModel(
                input_chunk_length=in_len,
                output_chunk_length=out_len,
                batch_size=8,
                n_epochs=2,
                nr_epochs_val_period=1,
                kernel_size=kernel_size,
                num_filters=num_filters,
                optimizer_kwargs={"lr": lr},
                add_encoders=encoders,
                likelihood=GaussianLikelihood(),
                pl_trainer_kwargs={
                    **tfm_kwargs["pl_trainer_kwargs"],
                    "callbacks": [pruner, early_stopper],
                },
                model_name="tcn_model",
                force_reset=True,
                save_checkpoints=True,
                work_dir=os.getcwd(),
            )

            # when validating during training, we can use a slightly longer validation
            # set which also contains the first input_chunk_length time steps
            model_val_set = self.scaler.transform(
                self.series[-(self.val_length + in_len) :]
            )

            # train the model
            model.fit(
                series=self.train,
                val_series=model_val_set,
            )

            # reload best model over course of training
            model = TCNModel.load_from_checkpoint(
                model_name="tcn_model", work_dir=os.getcwd()
            )

            # Evaluate how good it is on the validation set, using sMAPE
            preds = model.predict(series=self.train, n=self.val_length)
            smapes = smape(self.val, preds, n_jobs=-1)
            smape_val = np.mean(smapes)

            return smape_val if smape_val != np.nan else float("inf")

        # optimize hyperparameters by minimizing the sMAPE on the validation set
        study = optuna.create_study(direction="minimize")
        study.optimize(objective, n_trials=3)

    @pytest.mark.parametrize(
        "params",
        product(
            [True, False],  # multi_models
            [1, 3],  # ocl
        ),
    )
    def test_optuna_regression_model(self, params):
        """Check that optuna works as expected with a regression model"""

        multi_models, ocl = params

        # define objective function
        def objective(trial):
            # select input and encoder usage
            target_lags = trial.suggest_int("lags", 1, 12)
            include_year = trial.suggest_categorical("year", [False, True])

            # optionally also add the (scaled) year value as a past covariate
            if include_year:
                encoders = {
                    "datetime_attribute": {"past": ["year"]},
                    "transformer": Scaler(),
                }
                past_lags = trial.suggest_int("lags_past_covariates", 1, 12)
            else:
                encoders = None
                past_lags = None

            # build the model
            model = LinearRegressionModel(
                lags=target_lags,
                lags_past_covariates=past_lags,
                output_chunk_length=ocl,
                multi_models=multi_models,
                add_encoders=encoders,
            )
            model.fit(
                series=self.train,
            )

            # Evaluate how good it is on the validation set, using sMAPE
            preds = model.predict(series=self.train, n=self.val_length)
            smapes = smape(self.val, preds, n_jobs=-1)
            smape_val = np.mean(smapes)

            return smape_val if smape_val != np.nan else float("inf")

        # optimize hyperparameters by minimizing the sMAPE on the validation set
        study = optuna.create_study(direction="minimize")
        study.optimize(objective, n_trials=3)



"""
Interquartile Range (IQR) Detector
-----------------

Flags anomalies that are beyond the IQR (between the third and the first quartile)
of historical data by some factor of it's difference (typically 1.5).
This is similar to a threshold-based detector, but the thresholds are
computed as distances from the IQR of historical data when the detector is fitted.
"""

from collections.abc import Sequence
from typing import Union

import numpy as np

from darts.ad.detectors.quantile_detector import QuantileDetector
from darts.ad.detectors.threshold_detector import ThresholdDetector
from darts.logging import get_logger, raise_log
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class IQRDetector(QuantileDetector):
    def __init__(self, scale: Union[Sequence[float], float] = 1.5) -> None:
        """IQR Detector

        Flags values that lie outside of the interquartile range (IQR)
        by more than a certain factor of IQR's value as anomalies.
        The factor is passed in the `scale` parameter.

        If a single value is provided for `scale`,
        this same value will be used across all components of the series.

        If a sequences of values is given for the `scale` parameter,
        it's length must match the dimensionality of the series passed.

        Parameters
        ----------
        scale
            (Sequence of) scale(s) used to indicate what distance from the IQR constitutes an anomaly.
            Defaults to `1.5`. Must be non-negative. If a sequence, must match the dimensionality of the series
            this detector is applied to.
        """

        # Parent QuantileDetector will compute Q1 and Q3 thresholds
        super().__init__(low_quantile=0.25, high_quantile=0.75)

        self.scale = np.array(scale)
        if self.scale.ndim == 0:
            self.scale = np.expand_dims(self.scale, 0)

        if not np.issubdtype(self.scale.dtype, np.number) or (self.scale < 0.0).any():
            raise_log(
                ValueError("All values in `scale` must be non-negative numbers."),
                logger=logger,
            )

    def _fit_core(self, series: Sequence[TimeSeries]) -> None:
        super()._fit_core(series)

        if len(self.scale) > 1 and len(self.scale) != series[0].width:
            raise_log(
                ValueError(
                    "The number of components of input must be equal to the number "
                    "of values given for `scale`. Found number of components "
                    f"equal to {series[0].width} and expected {len(self.scale)}."
                ),
                logger=logger,
            )

        low_threshold = np.array(self.detector.low_threshold)
        high_threshold = np.array(self.detector.high_threshold)

        IQR = high_threshold - low_threshold

        low_threshold -= self.scale * IQR
        high_threshold += self.scale * IQR

        self.detector = ThresholdDetector(
            low_threshold=list(low_threshold), high_threshold=list(high_threshold)
        )



import numpy as np
import pandas as pd

from darts.dataprocessing.transformers import MissingValuesFiller
from darts.timeseries import TimeSeries


class TestMissingValuesFiller:
    time = pd.date_range("20130101", "20130130")
    static_covariate = pd.DataFrame({"0": [1]})

    const_series = TimeSeries.from_times_and_values(
        time, np.array([2.0] * len(time)), static_covariates=static_covariate
    )
    const_series_with_holes = TimeSeries.from_times_and_values(
        time,
        np.array([2.0] * 10 + [np.nan] * 5 + [2.0] * 10 + [np.nan] * 5),
        static_covariates=static_covariate,
    )

    lin = [float(i) for i in range(len(time))]
    lin_series = TimeSeries.from_times_and_values(time, np.array(lin))
    lin_series_with_holes = TimeSeries.from_times_and_values(
        time, np.array(lin[0:10] + [np.nan] * 5 + lin[15:24] + [np.nan] * 5 + [lin[29]])
    )

    def test_fill_const_series_with_const_value(self):
        const_transformer = MissingValuesFiller(fill=2.0)
        transformed = const_transformer.transform(self.const_series_with_holes)
        assert self.const_series == transformed

    def test_fill_const_series_with_auto_value(self):
        auto_transformer = MissingValuesFiller()
        transformed = auto_transformer.transform(self.const_series_with_holes)
        assert self.const_series == transformed

    def test_fill_lin_series_with_auto_value(self):
        auto_transformer = MissingValuesFiller()
        transformed = auto_transformer.transform(self.lin_series_with_holes)
        assert self.lin_series == transformed

    def test_fill_static_covariates_preserved(self):
        const_transformer = MissingValuesFiller(fill=2.0)
        transformed = const_transformer.transform(self.const_series_with_holes)
        assert (
            self.const_series.static_covariates.values
            == transformed.static_covariates.values
        )



"""
Hierarchical Reconciliation
---------------------------

A set of posthoc hierarchical reconciliation transformers. These transformers
work on any ``TimeSeries`` (e.g., a forecast) that contain a ``hierarchy``.

A ``hierarchy`` is a dict that maps each component to their parent(s) in the hierarchy.
It can be added to a ``TimeSeries`` using e.g., the :meth:`TimeSeries.with_hierarchy` method.
"""

from collections.abc import Mapping
from typing import Any, Optional

import numpy as np

from darts.dataprocessing.transformers import (
    BaseDataTransformer,
    FittableDataTransformer,
)
from darts.logging import get_logger, raise_if_not
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


def _get_summation_matrix(series: TimeSeries):
    """
    Returns the matrix S for a series, as defined `here <https://otexts.com/fpp3/reconciliation.html>`_.

    The dimension of the matrix is `(n, m)`, where `n` is the number of components and `m` the number
    of base components (components that are not the sum of any other components).
    S[i, j] contains 1 if component i is "made up" of base component j, and 0 otherwise.
    The order of the `n` and `m` components in the matrix match the order of the components in the `series`.

    The matrix is built using the ``hierarchy`` property of the ``series``. ``hierarchy`` must be a
    dictionary mapping each (non top-level) component to its parent(s) in the aggregation.
    """

    raise_if_not(
        series.has_hierarchy,
        "The provided series must have a hierarchy defined for reconciliation to be performed.",
        logger=logger,
    )
    hierarchy = series.hierarchy
    components_seq = list(series.components)
    leaves_seq = series.bottom_level_components
    m = len(leaves_seq)
    n = len(components_seq)
    S = np.zeros((n, m))

    components_indexes = {comp: i for i, comp in enumerate(components_seq)}
    leaves_indexes = {leaf: i for i, leaf in enumerate(leaves_seq)}

    def increment(cur_node, leaf_idx):
        """
        Recursive function filling S for a given base component and all its ancestors
        """
        S[components_indexes[cur_node], leaf_idx] = 1.0
        if cur_node in hierarchy:
            for parent in hierarchy[cur_node]:
                increment(parent, leaf_idx)

    for leaf in leaves_seq:
        leaf_idx = leaves_indexes[leaf]
        increment(leaf, leaf_idx)

    return S.astype(series.dtype)


def _reconcile_from_S_and_G(
    series: TimeSeries, S: np.ndarray, G: np.ndarray
) -> TimeSeries:
    """
    Returns the TimeSeries linearly reconciled from the projection matrix G and the summation matrix S.
    """
    y_hat = series.all_values(copy=False)
    reconciled_values = S @ G @ y_hat  # (n, m) * (m, n) * (time, n, samples)
    return series.with_values(reconciled_values)


class BottomUpReconciliator(BaseDataTransformer):
    """
    Performs bottom up reconciliation, as defined `here <https://otexts.com/fpp3/reconciliation.html>`_.
    """

    @staticmethod
    def get_projection_matrix(series):
        leaves_seq = list(series.bottom_level_components)
        n, m = series.n_components, len(leaves_seq)
        leaves_indexes = {leaf: i for i, leaf in enumerate(leaves_seq)}
        G = np.zeros((m, n))
        for i, c in enumerate(series.components):
            if c in leaves_indexes:
                G[leaves_indexes[c], i] = 1.0
        return G

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Any], *args, **kwargs
    ) -> TimeSeries:
        S = _get_summation_matrix(series)
        G = BottomUpReconciliator.get_projection_matrix(series)
        return _reconcile_from_S_and_G(series, S, G)


class TopDownReconciliator(FittableDataTransformer):
    """
    Performs top down reconciliation, as defined `here <https://otexts.com/fpp3/reconciliation.html>`_.

    This estimator computes the proportions (of the base components w.r.t. the top component)
    based on the TimeSeries provided to the method :meth:`fit()`. If the historical series
    is provided, then the historical proportions will be used.
    """

    @staticmethod
    def ts_fit(
        series: TimeSeries, params: Mapping[str, Any], *args, **kwargs
    ) -> np.ndarray:
        G = TopDownReconciliator.get_projection_matrix(series)
        return G

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Any], *args, **kwargs
    ) -> TimeSeries:
        G = params["fitted"]
        S = _get_summation_matrix(series)
        return _reconcile_from_S_and_G(series, S, G)

    @staticmethod
    def get_projection_matrix(series):
        n, m = series.n_components, len(series.bottom_level_components)

        # compute sum of total component
        sum_total = (
            series[series.top_level_component].all_values(copy=False).flatten().sum()
        )

        base_forecasts = series.bottom_level_series

        # compute sum of base components
        sum_base = base_forecasts.all_values(copy=False).sum(axis=2).sum(axis=0)

        # compute proportions for each base component
        proportions = sum_base / sum_total
        top_level_index = list(series.components).index(series.top_level_component)
        G = np.zeros((m, n))
        G[:, top_level_index] = proportions

        return G


class MinTReconciliator(FittableDataTransformer):
    def __init__(self, method="ols"):
        """
        MinT Reconcilator.

        This implements the MinT reconciliation approach presented in [1]_ and
        summarised in [2]_.

        Parameters
        ----------
        method
            This parameter can take four different values, determining how the covariance
            matrix ``W`` of the forecast errors is estimated (corresponding to ``Wh`` in [2]_):

            * ``ols`` uses ``W = I``. This option looks only at the hierarchy but ignores the
              values of the series provided to ``fit()``.
            * ``wls_struct`` uses ``W = diag(S1)``, where ``S1`` is a vector of size `n` with values
              between 0 and `m`, representing the number of base components composing each
              of the `n` components. This options looks only at the hierarchy but ignores
              the values of the series provided to ``fit()``.
            * ``wls_var`` uses ``W = diag(W1)``, where ``W1`` is the temporal average of the
              variance of the forecasting residuals. This method assumes that the series
              provided to ``fit()`` contain the forecast residuals (deterministic series).
            * ``mint_cov`` computes ``W`` as the empirical covariance matrix of the residuals
              for each component, with residuals samples taken over time. This method assumes
              that the series provided to ``fit()`` contain the forecast residuals
              (deterministic series), and it requires the residuals to be linearly independent.
            * ``wls_val`` uses ``W = diag(V1)``, where ``V1`` is the temporal average of the
              component values. This method assumes that the series provided to ``fit()`` contains
              an example of the actual values (e.g., either the training series or the forecasts).
              This method is not presented in [2]_.

        References
        ----------
        .. [1] `Optimal forecast reconciliation for hierarchical and grouped time series through
                trace minimization <https://robjhyndman.com/papers/MinT.pdf>`_
        .. [2] https://otexts.com/fpp3/reconciliation.html#the-mint-optimal-reconciliation-approach
        """
        known_methods = ["ols", "wls", "wls_var", "wls_struct", "wls_val", "mint_cov"]
        raise_if_not(
            method in known_methods,
            f"The method must be one of {known_methods}",
        )
        # Define fixed params (i.e. attributes defined before calling `super().__init__`):
        self.method = method
        super().__init__()

    @staticmethod
    def ts_fit(
        series: TimeSeries, params: Mapping[str, Any], *args, **kwargs
    ) -> np.ndarray:
        method = params["fixed"]["method"]
        S, G = MinTReconciliator.get_matrices(series, method)
        return S, G

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Any], *args, **kwargs
    ) -> TimeSeries:
        S, G = params["fitted"]
        return _reconcile_from_S_and_G(series, S, G)

    @staticmethod
    def _assert_deterministic(series: TimeSeries):
        raise_if_not(
            series.is_deterministic,
            "When used with method wls_var or mint_cov, the MinT reconciliator "
            + "has to be fit on a deterministic series "
            + "containing residuals. This series is stochastic.",
        )

    @staticmethod
    def get_matrices(series: Optional[TimeSeries], method: str):
        """Returns the G matrix given a specified reconciliation method."""
        S = _get_summation_matrix(series)
        if method == "ols":
            # G = inv(S'*S)*S'
            G = np.linalg.inv(S.T @ S) @ S.T
            return S, G
        elif method == "wls_struct":
            # Wh is a diagonal matrix with entry i,i being the sum of row i of S_mat
            Wh = np.diag(np.sum(S, axis=1))
        elif method == "wls_var":
            # In this case we assume that series contains the residuals of some forecasts
            MinTReconciliator._assert_deterministic(series)
            et2 = series.values(copy=False) ** 2  # squared residuals
            # Wh diagonal is mean squared residual over time:
            Wh = np.diag(et2.mean(axis=0))
        elif method == "wls_val":
            # Wh is a diagonal matrix with entry i,i being the average value of the corresponding time series
            quantities = series.all_values(copy=False).mean(axis=2).mean(axis=0)
            Wh = np.diag(np.array(quantities))
        elif method == "mint_cov":
            MinTReconciliator._assert_deterministic(series)
            Wh = np.cov(
                series.values(copy=False).T
            )  # + 1e-3 * np.eye(len(series.components))
        else:
            raise_if_not(False, f"Unknown method: {method}")

        Wh_inv = np.linalg.inv(Wh)
        G = np.linalg.inv(S.T @ Wh_inv @ S) @ S.T @ Wh_inv
        return S, G



"""
Utils for Anomaly Detection
---------------------------

Common functions used throughout the Anomaly Detection module.
"""

# TODO:
#     - migrate metrics function to darts.metric
#     - check error message
#     - create a zoom option on anomalies for a show function
#     - add an option to visualize: "by window", "unique", "together"
#     - create a normalize option in plot function (norm every anomaly score btw 1 and 0) -> to be seen on the same plot

from collections.abc import Sequence
from typing import Callable, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
)

from darts import TimeSeries
from darts.logging import get_logger, raise_log
from darts.utils.ts_utils import series2seq

logger = get_logger(__name__)


def eval_metric_from_scores(
    anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    pred_scores: Union[TimeSeries, Sequence[TimeSeries]],
    window: Union[int, Sequence[int]] = 1,
    metric: Literal["AUC_ROC", "AUC_PR"] = "AUC_ROC",
) -> Union[float, Sequence[float], Sequence[Sequence[float]]]:
    """Computes a score/metric between anomaly scores against true anomalies.

    `anomalies` and `pred_scores` must have the same shape.
    `anomalies` must be binary and have values belonging to the two classes (0 and 1).

    If one series is given for `anomalies` and `pred_scores` contains more than
    one series, the function will consider `anomalies` as the ground truth anomalies for
    all scores in `pred_scores`.

    Parameters
    ----------
    anomalies
        The (sequence of) ground truth binary anomaly series (`1` if it is an anomaly and `0` if not).
    pred_scores
        The (sequence of) of estimated anomaly score series indicating how anomalous each window of size w is.
    window
        Integer value indicating the number of past samples each point represents in the `pred_scores`.
        The parameter will be used to transform `anomalies`.
        If a list of integers, the length must match the number of series in `pred_scores`.
        If an integer, the value will be used for every series in `pred_scores` and `anomalies`.
    metric
        The name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
        Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
        Default: "AUC_ROC".

    Returns
    -------
    float
        A single score/metric for univariate `pred_scores` series (with only one component/column).
    Sequence[float]
        A sequence (list) of scores for:

        - multivariate `pred_scores` series (multiple components). Gives a score for each component.
        - a sequence (list) of univariate `pred_scores` series. Gives a score for each series.
    Sequence[Sequence[float]]
        A sequence of sequences of scores for a sequence of multivariate `pred_scores` series.
        Gives a score for each series (outer sequence) and component (inner sequence).
    """
    return _eval_metric(
        anomalies=anomalies,
        pred_series=pred_scores,
        window=window,
        metric=metric,
        pred_is_binary=False,
    )


def eval_metric_from_binary_prediction(
    anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    pred_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    window: Union[int, Sequence[int]] = 1,
    metric: Literal["recall", "precision", "f1", "accuracy"] = "recall",
) -> Union[float, Sequence[float], Sequence[Sequence[float]]]:
    """Computes a score/metric between predicted anomalies against true anomalies.

    `pred_anomalies` and `anomalies` must have:

        - identical dimensions (number of time steps and number of components/columns),
        - binary values belonging to the two classes (`1` if it is an anomaly and `0` if not)

    If one series is given for `anomalies` and `pred_anomalies` contains more than
    one series, the function will consider `anomalies` as the true anomalies for
    all scores in `pred_scores`.

    Parameters
    ----------
    anomalies
        The (sequence of) ground truth binary anomaly series (`1` if it is an anomaly and `0` if not).
    pred_anomalies
        The (sequence of) predicted binary anomaly series.
    window
        Integer value indicating the number of past samples each point represents in the `pred_scores`.
        The parameter will be used to transform `anomalies`.
        If a list of integers, the length must match the number of series in `pred_scores`.
        If an integer, the value will be used for every series in `pred_scores` and `anomalies`.
    metric
        The name of the metric function to use. Must be one of "recall", "precision", "f1", and "accuracy".
        Default: "recall".

    Returns
    -------
    float
        A single score for univariate `pred_anomalies` series (with only one component/column).
    Sequence[float]
        A sequence (list) of scores for:

        - multivariate `pred_anomalies` series (multiple components). Gives a score for each component.
        - a sequence (list) of univariate `pred_anomalies` series. Gives a score for each series.
    Sequence[Sequence[float]]
        A sequence of sequences of scores for a sequence of multivariate `pred_anomalies` series.
        Gives a score for each series (outer sequence) and component (inner sequence).
    """
    return _eval_metric(
        anomalies=anomalies,
        pred_series=pred_anomalies,
        window=window,
        metric=metric,
        pred_is_binary=True,
    )


def _eval_metric(
    anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    pred_series: Union[TimeSeries, Sequence[TimeSeries]],
    window: Union[int, Sequence[int]],
    metric: Literal["AUC_ROC", "AUC_PR", "recall", "precision", "f1", "accuracy"],
    pred_is_binary: bool,
):
    """Computes a score/metric between anomaly scores or binary predicted anomalies against true
    anomalies.

    Parameters
    ----------
    anomalies
        The (sequence of) ground truth binary anomaly series (`1` if it is an anomaly and `0` if not).
    pred_series
        The (sequence of) anomaly scores or predicted binary anomaly series.
    window
        Integer value indicating the number of past samples each point represents in the `pred_scores`.
        The parameter will be used to transform `anomalies`.
        If a list of integers, the length must match the number of series in `pred_scores`.
        If an integer, the value will be used for every series in `pred_scores` and `anomalies`.
    metric
        The name of the scoring function to use. Must be one of "recall", "precision",
        "f1", and "accuracy" if `pred_is_binary` is `True`. Otherwise, must be one of "AUC_ROC", "AUC_PR".
    pred_is_binary
        Whether `pred_series` refers predicted binary anomalies or anomaly scores.

    Returns
    -------
    float
        A single score for univariate `pred_series` series (with only one component/column).
    Sequence[float]
        A sequence (list) of scores for:

        - multivariate `pred_series` series (multiple components). Gives a score for each component.
        - a sequence (list) of univariate `pred_series` series. Gives a score for each series.
    Sequence[Sequence[float]]
        A sequence of sequences of scores for a sequence of multivariate `pred_series` series.
        Gives a score for each series (outer sequence) and component (inner sequence).
    """
    metrics_exp = (
        {"recall", "precision", "f1", "accuracy"}
        if pred_is_binary
        else {"AUC_ROC", "AUC_PR"}
    )
    if metric not in metrics_exp:
        raise_log(
            ValueError(f"Argument `metric` must be one of {metrics_exp}"),
            logger=logger,
        )

    if metric == "AUC_ROC":
        metric_fn = roc_auc_score
    elif metric == "AUC_PR":
        metric_fn = average_precision_score
    elif metric == "recall":
        metric_fn = recall_score
    elif metric == "precision":
        metric_fn = precision_score
    elif metric == "f1":
        metric_fn = f1_score
    else:
        metric_fn = accuracy_score

    called_with_single_series = isinstance(pred_series, TimeSeries)
    anomalies = series2seq(anomalies)
    pred_series = series2seq(pred_series)
    window = [window] if not isinstance(window, Sequence) else window

    if len(anomalies) == 1 and len(pred_series) > 1:
        anomalies = anomalies * len(pred_series)

    name = "anomalies"
    pred_name = "pred_anomalies" if pred_is_binary else "pred_scores"
    _assert_same_length(
        anomalies,
        pred_series,
        name,
        pred_name,
    )

    if len(window) == 1:
        window = window * len(anomalies)
    else:
        if len(window) != len(anomalies):
            raise_log(
                ValueError(
                    f"The list of windows must be the same length as the list of `{pred_name}` and "
                    f"`{name}`. There must be one window value for each series. "
                    f"Found length {len(window)}, expected {len(anomalies)}."
                ),
                logger=logger,
            )

    sol = []
    for s_anomalies, s_pred, s_window in zip(anomalies, pred_series, window):
        _assert_timeseries(s_pred, name=pred_name)
        _assert_timeseries(s_anomalies, name=name)
        _assert_binary(s_anomalies, name)
        if pred_is_binary:
            _assert_binary(s_pred, pred_name)

        # if s_window > 1, the anomalies will be adjusted so that it can be compared timewise with s_pred
        s_anomalies = _max_pooling(s_anomalies, s_window)

        _sanity_check_two_series(s_pred, s_anomalies, pred_name, name)

        s_pred_vals = s_pred.slice_intersect_values(s_anomalies, copy=False)
        s_anomalies_vals = s_anomalies.slice_intersect_values(s_pred, copy=False)

        if not len(s_pred_vals) == len(s_anomalies_vals):
            raise_log(
                ValueError(
                    f"The two time series `{pred_name}` and `{name}` "
                    f"must have at least a partially overlapping time index."
                ),
                logger=logger,
            )

        if not pred_is_binary:  # `pred_series` is an anomaly score
            nr_anomalies_per_component = s_anomalies_vals.sum(axis=0).flatten()

            if nr_anomalies_per_component.min() == 0:
                raise_log(
                    ValueError(
                        f"`{name}` does not contain anomalies. {metric} cannot be computed."
                    ),
                    logger=logger,
                )
            if nr_anomalies_per_component.max() == len(s_anomalies_vals):
                add_txt = (
                    ""
                    if s_window <= 1
                    else f" Consider decreasing the window size (window={s_window})"
                )
                raise_log(
                    ValueError(
                        f"`{name}` only contains anomalies. {metric} cannot be computed."
                        + add_txt
                    ),
                    logger=logger,
                )

        # TODO: could we vectorize this?
        metrics = []
        for component_idx in range(s_pred.width):
            metrics.append(
                metric_fn(
                    s_anomalies_vals[:, component_idx],
                    s_pred_vals[:, component_idx],
                )
            )
        sol.append(metrics if len(metrics) > 1 else metrics[0])

    return sol[0] if called_with_single_series else sol


def show_anomalies_from_scores(
    series: TimeSeries,
    anomalies: TimeSeries = None,
    pred_series: TimeSeries = None,
    pred_scores: Union[TimeSeries, Sequence[TimeSeries]] = None,
    window: Union[int, Sequence[int]] = 1,
    names_of_scorers: Union[str, Sequence[str]] = None,
    title: str = None,
    metric: Optional[Literal["AUC_ROC", "AUC_PR"]] = None,
    component_wise: bool = False,
):
    """Plot the results generated by an anomaly model.

    The plot will be composed of the following:
        - the actual series itself with the output of the model (if given)
        - the anomaly score of each scorer. The scorer with different windows will be separated.
        - the actual anomalies, if given.

    If `pred_series` is stochastic (i.e., if it has multiple samples), the function will plot:
        - the mean per timestamp
        - the quantile 0.95 for an upper bound
        - the quantile 0.05 for a lower bound

    Possible to:
        - add a title to the figure with the parameter `title`
        - give personalized names for the scorers with `names_of_scorers`
        - show the results of a metric for each anomaly score (AUC_ROC or AUC_PR), if the actual anomalies is given

    Parameters
    ----------
    series
        The actual series to visualize anomalies from.
    anomalies
        The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
    pred_series
        Output of the model given as input the `series` (can be stochastic).
    pred_scores
        Output of the scorers given the output of the model and `series`.
    window
        Window parameter for each anomaly scores.
        Default: 1. If a list of anomaly scores is given, the same default window will be used for every score.
    names_of_scorers
        Name of the scores. Must be a list of length equal to the number of scorers in the anomaly_model.
        Only effective when `pred_scores` is not `None`.
    title
        Title of the figure
    metric
        Optionally, the name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
        Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
        Only effective when `pred_scores` is not `None`.
        Default: "AUC_ROC".
    component_wise
        If True, will separately plot each component in case of multivariate anomaly detection.
    """
    series = _check_input(
        series,
        name="series",
        num_series_expected=1,
        check_multivariate=component_wise,
    )[0]

    if title is None and pred_scores is not None:
        title = "Anomaly results"

    nbr_plots = 1
    if anomalies is not None:
        nbr_plots = nbr_plots + 1
    elif metric is not None:
        raise_log(
            ValueError("`anomalies` must be given in order to calculate a metric."),
            logger=logger,
        )

    pred_scores = series2seq(pred_scores)
    if pred_scores is not None:
        if names_of_scorers is not None:
            names_of_scorers = (
                [names_of_scorers]
                if isinstance(names_of_scorers, str)
                else names_of_scorers
            )
            if len(names_of_scorers) != len(pred_scores):
                raise_log(
                    ValueError(
                        f"The number of names in `names_of_scorers` must match the "
                        f"number of anomaly score given as input, found "
                        f"{len(names_of_scorers)} and expected {len(pred_scores)}."
                    ),
                    logger=logger,
                )

        window = [window] if isinstance(window, int) else window
        if not all([w > 0 for w in window]):
            raise_log(
                ValueError(
                    "Parameter `window` must be a positive integer, "
                    "or a sequence of positive integers."
                ),
                logger=logger,
            )
        window = window if len(window) > 1 else window * len(pred_scores)
        if len(window) != len(pred_scores):
            raise_log(
                ValueError(
                    f"The number of window in `window` must match the "
                    f"number of anomaly score given as input. One window "
                    f"value for each series. Found length {len(window)}, "
                    f"and expected {len(pred_scores)}."
                ),
                logger=logger,
            )

        if not all([w < len(s) for (w, s) in zip(window, pred_scores)]):
            raise_log(
                ValueError(
                    "Parameter `window` must be an integer or sequence of integers "
                    "with value(s) smaller than the length of the corresponding series "
                    "in `pred_scores`."
                ),
                logger=logger,
            )

        nbr_plots += len(set(window))

    series_width = series.n_components
    if pred_series is not None:
        pred_series = _check_input(
            pred_series,
            name="pred_series",
            width_expected=series_width,
            num_series_expected=1,
            check_multivariate=component_wise,
        )[0]

    if anomalies is not None and component_wise:
        anomalies = _check_input(
            anomalies,
            name="anomalies",
            width_expected=series_width,
            num_series_expected=1,
            check_binary=True,
            check_multivariate=component_wise,
        )[0]

    if pred_scores is not None and component_wise:
        for pred_score in pred_scores:
            _ = _check_input(
                pred_score,
                name="pred_score",
                width_expected=series_width,
                num_series_expected=1,
                check_multivariate=component_wise,
            )[0]

    plots_per_ts = nbr_plots * series_width if component_wise else nbr_plots
    height_ratios = ([2] + [1] * (nbr_plots - 1)) * (plots_per_ts // nbr_plots)
    height_total = 2 * sum(height_ratios)
    fig, axs = plt.subplots(
        nrows=plots_per_ts,
        figsize=(8, height_total),
        sharex=True,
        gridspec_kw={"height_ratios": height_ratios},
    )

    for i in range(series_width if component_wise else 1):
        if component_wise:
            series_ = series[series.components[i]]
            anomalies_ = (
                anomalies[anomalies.components[i]] if anomalies is not None else None
            )
            pred_series_ = (
                pred_series[pred_series.components[i]]
                if pred_series is not None
                else None
            )
            pred_scores_ = (
                [pc[pc.components[i]] for pc in pred_scores]
                if pred_scores is not None
                else None
            )
        else:
            series_ = series
            anomalies_ = anomalies
            pred_series_ = pred_series
            pred_scores_ = pred_scores

        _plot_series_and_anomalies(
            series=series_,
            anomalies=anomalies_,
            pred_series=pred_series_,
            pred_scores=pred_scores_,
            window=window,
            names_of_scorers=names_of_scorers,
            metric=metric,
            axs=axs,
            index_ax=i * nbr_plots,
        )
    # make title fit nicely on plot
    title_height = 0.1
    title_y = 1 - title_height / height_total

    fig.suptitle(title, y=title_y)
    fig.tight_layout()


def _assert_binary(series: TimeSeries, name: str):
    """Checks if series is a binary timeseries (1 and 0)"

    Parameters
    ----------
    series
        series to check for.
    name
        name of the series.
    """

    vals = series.values(copy=False)
    if not np.array_equal(vals, vals.astype(bool)):
        raise_log(
            ValueError(f"Input series `{name}` must have binary values only."),
            logger=logger,
        )


def _assert_timeseries(series: TimeSeries, name: str = "series"):
    """Checks if given input is of type Darts TimeSeries"""
    if not isinstance(series, TimeSeries):
        raise_log(
            ValueError(
                f"all series in `{name}` must be `TimeSeries`. Received {type(series)}."
            ),
            logger=logger,
        )


def _sanity_check_two_series(
    series_1: TimeSeries,
    series_2: TimeSeries,
    name_series_1: str,
    name_series_2: str,
):
    """Performs sanity check on the two given inputs

    Checks if the two inputs:
        - type is Darts Timeseries
        - have the same number of components
        - if their intersection in time is not null

    Parameters
    ----------
    series_1
        1st time series
    series_2:
        2nd time series
    """

    _assert_timeseries(series_1, name=name_series_1)
    _assert_timeseries(series_2, name=name_series_2)

    # check if the two inputs time series have the same number of components
    if series_1.width != series_2.width:
        raise_log(
            ValueError(
                f"The series from `{name_series_1}` and `{name_series_2}` must have the "
                f"same number of components, found {series_1.width} and {series_2.width}."
            ),
            logger=logger,
        )


def _max_pooling(series: TimeSeries, window: int) -> TimeSeries:
    """Slides a window of size `window` along the input series, and replaces the value of the
    input time series by the maximum of the values contained in the window.

    The binary time series output represents if there is an anomaly (=1) or not (=0) in the past
    window points. The new series will equal the length of the input series - window. Its first
    point will start at the first time index of the input time series + window points.

    Parameters
    ----------
    series:
        Binary time series.
    window:
        Integer value indicating the number of past samples each point represents.

    Returns
    -------
    Binary TimeSeries
    """
    if window <= 0:
        raise_log(
            ValueError(
                f"Parameter `window` must be strictly greater than 0, found size {window}."
            ),
            logger=logger,
        )
    if window >= len(series):
        raise_log(
            ValueError(
                f"Parameter `window` must be smaller than the length of the "
                f"input series, found window size {window}, and max size {len(series)}."
            ),
            logger=logger,
        )

    if window == 1:
        # the process results in replacing every value by itself -> return directly the series
        return series

    return series.window_transform(
        transforms={
            "window": window,
            "function": "max",
            "mode": "rolling",
            "min_periods": window,
        },
        treat_na="dropna",
    )


def _assert_same_length(
    list_series_1: Sequence[TimeSeries],
    list_series_2: Sequence[TimeSeries],
    name_series_1: str,
    name_series_2: str,
):
    """Checks if the two sequences contain the same number of TimeSeries."""

    if len(list_series_1) != len(list_series_2):
        raise_log(
            ValueError(
                f"Number of `{name_series_2}` must match the number of given "
                f"`{name_series_1}`, found length {len(list_series_2)} and "
                f"expected {len(list_series_1)}."
            ),
            logger=logger,
        )


def _plot_series(series, ax_id, linewidth, label_name, **kwargs):
    """Internal function called by `show_anomalies_from_scores()`

    Plot the series on the given axes ax_id.

    Parameters
    ----------
    series
        The series to plot.
    ax_id
        The axis the series will be plotted on.
    linewidth
        Thickness of the line.
    label_name
        Name that will appear in the legend.
    """
    for i, c in enumerate(series._xa.component[:10]):
        comp = series._xa.sel(component=c)

        if comp.sample.size > 1:
            central_series = comp.mean(dim="sample")
            low_series = comp.quantile(q=0.05, dim="sample")
            high_series = comp.quantile(q=0.95, dim="sample")
        else:
            central_series = comp

        label_to_use = (
            (label_name + ("_" + str(i) if len(series.components) > 1 else ""))
            if label_name != ""
            else "" + str(str(c.values))
        )

        central_series.plot(ax=ax_id, linewidth=linewidth, label=label_to_use, **kwargs)

        if comp.sample.size > 1:
            ax_id.fill_between(
                series.time_index, low_series, high_series, alpha=0.25, **kwargs
            )


def _check_input(
    series: Union[TimeSeries, Sequence[TimeSeries]],
    name: str,
    width_expected: Optional[int] = None,
    check_deterministic: bool = False,
    check_binary: bool = False,
    check_multivariate: bool = False,
    num_series_expected: Optional[int] = None,
    extra_checks: Optional[Callable] = None,
):
    """
    Input `series` checks used for Aggregators, Detectors, ...

    - `series` must be (sequence of) series with length (`num_series_expected`) where each series must:
        - have width `width_expected` if it is not `None`
        - be deterministic if `check_deterministic=True`
        - be binary if `check_binary=True`
        - be multivariate if `check_multivariate=True`

    By default, all checks except the `TimeSeries` check are disabled.

    Parameters
    ----------
    series
        A (sequence of) multivariate series.
    name
        The name of the series.
    width_expected
        Optionally, the expected number of components/width of each series.
    check_multivariate
        Whether to check if all series are multivariate.
    """
    series = series2seq(series)
    if num_series_expected is not None and len(series) != num_series_expected:
        if num_series_expected == 1:
            err_txt = f"`{name}` must be single `TimeSeries` or a sequence of `TimeSeries` of length `1`."
        else:
            err_txt = f"`{name}` must be a sequence of `TimeSeries` of length `{num_series_expected}`."
        raise_log(
            ValueError(err_txt),
            logger=logger,
        )
    for s in series:
        if not isinstance(s, TimeSeries):
            raise_log(
                ValueError(f"all series in `{name}` must be of type `TimeSeries`."),
                logger=logger,
            )
        if check_deterministic and not s.is_deterministic:
            raise_log(
                ValueError(
                    f"all series in `{name}` must be deterministic (number of samples=1)."
                ),
                logger=logger,
            )
        if check_binary:
            _assert_binary(s, name=name)
        if check_multivariate and s.width <= 1:
            raise_log(
                ValueError(f"all series in `{name}` must be multivariate (width>1)."),
                logger=logger,
            )
        if width_expected is not None and s.width != width_expected:
            raise_log(
                ValueError(
                    f"all series in `{name}` must have `{width_expected}` component(s) (width={width_expected})."
                ),
                logger=logger,
            )
        if extra_checks is not None:
            extra_checks(s)
    return series


def _assert_fit_called(fit_called: bool, name: str):
    """Checks that `fit_called` is `True`."""
    if not fit_called:
        raise_log(
            ValueError(
                f"The `{name}` has not been fitted yet. Call `{name}.fit()` first."
            ),
            logger=logger,
        )


def _plot_series_and_anomalies(
    series: TimeSeries,
    anomalies: TimeSeries,
    pred_series: TimeSeries,
    pred_scores: Sequence[TimeSeries],
    window: Sequence[int],
    names_of_scorers: Sequence[str],
    metric: str,
    axs: plt.Axes,
    index_ax: int,
):
    """Helper function to plot series and anomalies.

    Parameters
    ----------
    series
        The actual series to visualize anomalies from.
    anomalies
        The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
    pred_series
        Output of the model given as input the `series` (can be stochastic).
    pred_scores
        Output of the scorers given the output of the model and `series`.
    window
        Window parameter for each anomaly scores.
    names_of_scorers
        Name of the scores.
    metric
        The name of the metric function to use.
    axs
        The axes to plot on.
    index_ax
        The index of the current axis.
    """
    _plot_series(series=series, ax_id=axs[index_ax], linewidth=0.5, label_name="")

    if pred_series is not None:
        _plot_series(
            series=pred_series,
            ax_id=axs[index_ax],
            linewidth=0.5,
            label_name="model output",
        )

    axs[index_ax].set_title("")

    if anomalies is not None or pred_scores is not None:
        axs[index_ax].set_xlabel("")

    axs[index_ax].legend(loc="upper center", bbox_to_anchor=(0.5, 1.1), ncol=2)

    if pred_scores is not None:
        dict_input = {}

        for idx, (score, w) in enumerate(zip(pred_scores, window)):
            dict_input[idx] = {"series_score": score, "window": w, "name_id": idx}

        for index, elem in enumerate(
            sorted(dict_input.items(), key=lambda x: x[1]["window"])
        ):
            if index == 0:
                current_window = elem[1]["window"]
                index_ax = index_ax + 1

            idx = elem[1]["name_id"]
            w = elem[1]["window"]

            if w != current_window:
                current_window = w
                index_ax = index_ax + 1

            if metric is not None:
                value = round(
                    eval_metric_from_scores(
                        anomalies=anomalies,
                        pred_scores=pred_scores[idx],
                        window=w,
                        metric=metric,
                    ),
                    3,
                )
            else:
                value = None

            if names_of_scorers is not None:
                label = names_of_scorers[idx] + [f" ({value})", ""][value is None]
            else:
                label = f"score_{str(idx)}" + [f" ({value})", ""][value is None]

            _plot_series(
                series=elem[1]["series_score"],
                ax_id=axs[index_ax],
                linewidth=0.5,
                label_name=label,
            )

            axs[index_ax].legend(loc="upper center", bbox_to_anchor=(0.5, 1.19), ncol=2)
            axs[index_ax].set_title(f"Window: {str(w)}", loc="left")
            axs[index_ax].set_title("")
            axs[index_ax].set_xlabel("")

    if anomalies is not None:
        _plot_series(
            series=anomalies,
            ax_id=axs[index_ax + 1],
            linewidth=1,
            label_name="anomalies",
            color="red",
        )

        axs[index_ax + 1].set_title("")
        axs[index_ax + 1].set_ylim([-0.1, 1.1])
        axs[index_ax + 1].set_yticks([0, 1])
        axs[index_ax + 1].set_yticklabels(["no", "yes"])
        axs[index_ax + 1].legend(loc="upper center", bbox_to_anchor=(0.5, 1.2), ncol=2)
    else:
        axs[index_ax].set_xlabel("timestamp")



from collections.abc import Sequence
from itertools import product

import numpy as np
import pandas as pd
import pytest
from pyod.models.knn import KNN

from darts import TimeSeries
from darts.ad import (
    AndAggregator,  # noqa: F401
    CauchyNLLScorer,
    EnsembleSklearnAggregator,  # noqa: F401
    ExponentialNLLScorer,
    FilteringAnomalyModel,
    ForecastingAnomalyModel,
    GammaNLLScorer,
    GaussianNLLScorer,
    KMeansScorer,
    LaplaceNLLScorer,
    OrAggregator,  # noqa: F401
    PoissonNLLScorer,
    PyODScorer,
    QuantileDetector,  # noqa: F401
    ThresholdDetector,  # noqa: F401
    WassersteinScorer,
)
from darts.ad import DifferenceScorer as Difference
from darts.ad import NormScorer as Norm
from darts.ad.utils import eval_metric_from_scores, show_anomalies_from_scores
from darts.models import MovingAverageFilter, NaiveSeasonal, RegressionModel

filtering_am = [
    (
        FilteringAnomalyModel,
        {"model": MovingAverageFilter(window=10), "scorer": Norm()},
    ),
    (
        FilteringAnomalyModel,
        {"model": MovingAverageFilter(window=10), "scorer": [Norm(), KMeansScorer()]},
    ),
    (
        FilteringAnomalyModel,
        {"model": MovingAverageFilter(window=10), "scorer": KMeansScorer()},
    ),
]

forecasting_am = [
    (ForecastingAnomalyModel, {"model": RegressionModel(lags=10), "scorer": Norm()}),
    (
        ForecastingAnomalyModel,
        {"model": RegressionModel(lags=10), "scorer": [Norm(), KMeansScorer()]},
    ),
    (
        ForecastingAnomalyModel,
        {"model": RegressionModel(lags=10), "scorer": KMeansScorer()},
    ),
]


class TestAnomalyDetectionModel:
    np.random.seed(42)

    # univariate series
    np_train = np.random.normal(loc=10, scale=0.5, size=100)
    train = TimeSeries.from_values(np_train)

    np_covariates = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
    covariates = TimeSeries.from_times_and_values(train._time_index, np_covariates)

    np_test = np.random.normal(loc=10, scale=1, size=100)
    test = TimeSeries.from_times_and_values(train._time_index, np_test)

    np_anomalies = np.random.choice(a=[0, 1], size=100, p=[0.9, 0.1])
    anomalies = TimeSeries.from_times_and_values(train._time_index, np_anomalies)

    np_only_1_anomalies = np.random.choice(a=[0, 1], size=100, p=[0, 1])
    only_1_anomalies = TimeSeries.from_times_and_values(
        train._time_index, np_only_1_anomalies
    )

    np_only_0_anomalies = np.random.choice(a=[0, 1], size=100, p=[1, 0])
    only_0_anomalies = TimeSeries.from_times_and_values(
        train._time_index, np_only_0_anomalies
    )

    modified_train = MovingAverageFilter(window=10).filter(train)
    modified_test = MovingAverageFilter(window=10).filter(test)

    np_probabilistic = np.random.normal(loc=10, scale=1, size=[100, 1, 20])
    probabilistic = TimeSeries.from_times_and_values(
        train._time_index, np_probabilistic
    )

    # multivariate series
    np_mts_train = np.random.normal(loc=[10, 5], scale=[0.5, 1], size=[100, 2])
    mts_train = TimeSeries.from_values(np_mts_train)

    np_mts_test = np.random.normal(loc=[10, 5], scale=[1, 1.5], size=[100, 2])
    mts_test = TimeSeries.from_times_and_values(mts_train._time_index, np_mts_test)

    np_mts_anomalies = np.random.choice(a=[0, 1], size=[100, 2], p=[0.9, 0.1])
    mts_anomalies = TimeSeries.from_times_and_values(
        mts_train._time_index, np_mts_anomalies
    )

    @pytest.mark.parametrize(
        "scorer,anomaly_model_config",
        product(
            [
                Norm(),
                Difference(),
                GaussianNLLScorer(),
                ExponentialNLLScorer(),
                PoissonNLLScorer(),
                LaplaceNLLScorer(),
                CauchyNLLScorer(),
                GammaNLLScorer(),
            ],
            [
                (ForecastingAnomalyModel, {"model": RegressionModel(lags=10)}),
                (FilteringAnomalyModel, {"model": MovingAverageFilter(window=20)}),
            ],
        ),
    )
    def test_non_fittable_scorer(self, scorer, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(scorer=scorer, **am_kwargs)
        assert not anomaly_model.scorers_are_trainable

    @pytest.mark.parametrize(
        "scorer,anomaly_model_config",
        product(
            [
                PyODScorer(model=KNN()),
                KMeansScorer(),
                WassersteinScorer(window_agg=False),
            ],
            [
                (ForecastingAnomalyModel, {"model": RegressionModel(lags=10)}),
                (FilteringAnomalyModel, {"model": MovingAverageFilter(window=20)}),
            ],
        ),
    )
    def test_fittable_scorer(self, scorer, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(scorer=scorer, **am_kwargs)
        assert anomaly_model.scorers_are_trainable

    def test_no_local_model(self):
        with pytest.raises(ValueError) as err:
            _ = ForecastingAnomalyModel(model=NaiveSeasonal(), scorer=KMeansScorer())
        assert str(err.value) == "`model` must be a Darts `GlobalForecastingModel`."

    @pytest.mark.parametrize(
        "anomaly_model,fit_model",
        [
            (
                ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer=Norm()),
                True,
            ),
            (
                FilteringAnomalyModel(
                    model=MovingAverageFilter(window=20), scorer=Norm()
                ),
                False,
            ),
        ],
    )
    def test_score(self, anomaly_model, fit_model):
        if fit_model:
            anomaly_model.fit(self.train, allow_model_training=True)

        # if return_model_prediction set to true, output must be tuple
        assert isinstance(
            anomaly_model.score(self.test, return_model_prediction=True), tuple
        )

        # if return_model_prediction set to false output must be
        # Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
        assert not isinstance(
            anomaly_model.score(self.test, return_model_prediction=False), tuple
        )

    @pytest.mark.parametrize("anomaly_model_config", filtering_am)
    def test_FitFilteringAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)
        # `allow_model_training=True` has no effect if filter model has no `fit()` method
        anomaly_model.fit(self.train, allow_model_training=True)

        # input 'series' must be a series or Sequence of series
        with pytest.raises(ValueError):
            anomaly_model.fit([self.train, "str"], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([[self.train, self.train]], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit("str", allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([1, 2, 3], allow_model_training=True)

    @pytest.mark.parametrize("anomaly_model_config", forecasting_am)
    def test_FitForecastingAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)

        # input 'series' must be a series or Sequence of series
        with pytest.raises(ValueError):
            anomaly_model.fit([self.train, "str"], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([[self.train, self.train]], allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit("str", allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.fit([1, 2, 3], allow_model_training=True)

        # 'allow_model_training' must be set to True if forecasting model is not fitted
        if anomaly_model.scorers_are_trainable:
            with pytest.raises(ValueError):
                anomaly_model.fit(self.train, allow_model_training=False)
                anomaly_model.score(self.train)

        with pytest.raises(ValueError):
            # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=[self.train, self.train],
                past_covariates=self.covariates,
                allow_model_training=True,
            )

        with pytest.raises(ValueError):
            # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=self.train,
                past_covariates=[self.covariates, self.covariates],
                allow_model_training=True,
            )

        with pytest.raises(ValueError):
            # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=[self.train, self.train],
                future_covariates=self.covariates,
                allow_model_training=True,
            )

        with pytest.raises(ValueError):
            # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
            anomaly_model.fit(
                series=self.train,
                future_covariates=[self.covariates, self.covariates],
                allow_model_training=True,
            )

    def test_pretrain_forecasting_model(self):
        fitted_model = RegressionModel(lags=10).fit(self.train)
        # Fittable scorer must be fitted before calling .score(), even if forecasting model is fitted
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=fitted_model, scorer=KMeansScorer()).score(
                series=self.test
            )
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=fitted_model, scorer=[Norm(), KMeansScorer()]
            ).score(series=self.test)

        # forecasting model that do not accept past/future covariates
        # with pytest.raises(ValueError):
        #    ForecastingAnomalyModel(model=ExponentialSmoothing(),
        #       scorer=NormScorer()).fit(
        #           series=self.train, past_covariates=self.covariates, allow_model_training=True
        #       )
        # with pytest.raises(ValueError):
        #    ForecastingAnomalyModel(model=ExponentialSmoothing(),
        #       scorer=NormScorer()).fit(
        #           series=self.train, future_covariates=self.covariates, allow_model_training=True
        #       )

        # check window size
        # max window size is len(series.drop_before(series.get_timestamp_at_point(start))) + 1
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=RegressionModel(lags=10),
                scorer=KMeansScorer(window=50, window_agg=False),
            ).fit(series=self.train, start=0.9)

        # forecasting model that cannot be trained on a list of series
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=NaiveSeasonal(), scorer=Norm()).fit(
                series=[self.train, self.train], allow_model_training=True
            )

    @pytest.mark.parametrize("anomaly_model_config", forecasting_am)
    def test_ScoreForecastingAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)
        anomaly_model.fit(self.train, allow_model_training=True)

        # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=[self.train, self.train], past_covariates=self.covariates
            )

        # number of 'past_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=self.train,
                past_covariates=[self.covariates, self.covariates],
            )

        # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=[self.train, self.train], future_covariates=self.covariates
            )

        # number of 'future_covariates' must be the same as the number of Timeseries in 'series'
        with pytest.raises(ValueError):
            anomaly_model.score(
                series=self.train,
                future_covariates=[self.covariates, self.covariates],
            )

    def test_window_size(self):
        # max window size is len(series.drop_before(series.get_timestamp_at_point(start))) + 1 for score()
        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10),
            scorer=KMeansScorer(window=30, window_agg=False),
        )
        anomaly_model.fit(self.train, allow_model_training=True)
        with pytest.raises(ValueError):
            anomaly_model.score(series=self.train, start=0.9)

    @pytest.mark.parametrize("anomaly_model_config", filtering_am)
    def test_ScoreFilteringAnomalyModelInput(self, anomaly_model_config):
        am_cls, am_kwargs = anomaly_model_config
        anomaly_model = am_cls(**am_kwargs)

        if anomaly_model.scorers_are_trainable:
            anomaly_model.fit(self.train)

    @pytest.mark.parametrize(
        "anomaly_model,fit_kwargs",
        [
            (
                ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer=Norm()),
                {"series": train, "allow_model_training": True},
            ),
            (
                FilteringAnomalyModel(
                    model=MovingAverageFilter(window=20), scorer=Norm()
                ),
                False,
            ),
            (
                ForecastingAnomalyModel(
                    model=RegressionModel(lags=10),
                    scorer=[Norm(), WassersteinScorer(window_agg=False)],
                ),
                {"series": train, "allow_model_training": True},
            ),
            (
                FilteringAnomalyModel(
                    model=MovingAverageFilter(window=20),
                    scorer=[Norm(), WassersteinScorer(window_agg=False)],
                ),
                {"series": train},
            ),
        ],
    )
    def test_eval_metric(self, anomaly_model, fit_kwargs):
        if fit_kwargs:
            anomaly_model.fit(**fit_kwargs)

        # if the anomaly_model have scorers that have the parameter is_univariate set to True,
        # 'anomalies' must have widths of 1
        if anomaly_model.scorers_are_univariate:
            with pytest.raises(ValueError):
                anomaly_model.eval_metric(
                    anomalies=self.mts_anomalies, series=self.test
                )
            with pytest.raises(ValueError):
                anomaly_model.eval_metric(
                    anomalies=self.mts_anomalies, series=self.mts_test
                )
            with pytest.raises(ValueError):
                anomaly_model.eval_metric(
                    anomalies=[self.anomalies, self.mts_anomalies],
                    series=[self.test, self.mts_test],
                )

        # 'metric' must be str and "AUC_ROC" or "AUC_PR"
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies, series=self.test, metric=1
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies,
                series=self.test,
                metric="auc_roc",
            )
        with pytest.raises(TypeError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies,
                series=self.test,
                metric=["AUC_ROC"],
            )

        # 'anomalies' must be binary
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(anomalies=self.test, series=self.test)

        # 'anomalies' must contain anomalies (at least one)
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(anomalies=self.only_0_anomalies, series=self.test)

        # 'anomalies' cannot contain only anomalies
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(anomalies=self.only_1_anomalies, series=self.test)

        # 'anomalies' must match the number of series
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies, series=[self.test, self.test]
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=[self.anomalies, self.anomalies],
                series=self.test,
            )

        # 'anomalies' must have non empty intersection with 'series'
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=self.anomalies[:20], series=self.test[30:]
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                anomalies=[self.anomalies, self.anomalies[:20]],
                series=[self.test, self.test[40:]],
            )

        # Check input type
        # 'anomalies' and 'series' must be of same length
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies], [self.test, self.test])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(self.anomalies, [self.test, self.test])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies, self.anomalies], [self.test])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies, self.anomalies], self.test)

        # 'anomalies' and 'series' must be of type Timeseries
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies], [2, 3, 4])
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([self.anomalies], "str")
        with pytest.raises(ValueError):
            anomaly_model.eval_metric([2, 3, 4], self.test)
        with pytest.raises(ValueError):
            anomaly_model.eval_metric("str", self.test)
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                [self.anomalies, self.anomalies], [self.test, [3, 2, 1]]
            )
        with pytest.raises(ValueError):
            anomaly_model.eval_metric(
                [self.anomalies, [3, 2, 1]], [self.test, self.test]
            )

        # Check return types
        # Check if return type is float when input is a series
        assert isinstance(
            anomaly_model.eval_metric(self.anomalies, self.test),
            dict,
        )

        # Check if return type is Sequence when input is a Sequence of series
        assert isinstance(
            anomaly_model.eval_metric(self.anomalies, [self.test]),
            Sequence,
        )

        assert isinstance(
            anomaly_model.eval_metric(
                [self.anomalies, self.anomalies], [self.test, self.test]
            ),
            Sequence,
        )

    def test_ForecastingAnomalyModelInput(self):
        # model input
        # model input must be of type ForecastingModel
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model="str", scorer=Norm())
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=1, scorer=Norm())
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=MovingAverageFilter(window=10), scorer=Norm())
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=[RegressionModel(lags=10), RegressionModel(lags=5)],
                scorer=Norm(),
            )

        # scorer input
        # scorer input must be of type AnomalyScorer
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer=1)
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(model=RegressionModel(lags=10), scorer="str")
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=RegressionModel(lags=10), scorer=RegressionModel(lags=10)
            )
        with pytest.raises(ValueError):
            ForecastingAnomalyModel(
                model=RegressionModel(lags=10), scorer=[Norm(), "str"]
            )

    def test_FilteringAnomalyModelInput(self):
        # model input
        # model input must be of type FilteringModel
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model="str", scorer=Norm())
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=1, scorer=Norm())
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=RegressionModel(lags=10), scorer=Norm())
        with pytest.raises(ValueError):
            FilteringAnomalyModel(
                model=[MovingAverageFilter(window=10), MovingAverageFilter(window=10)],
                scorer=Norm(),
            )

        # scorer input
        # scorer input must be of type AnomalyScorer
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=MovingAverageFilter(window=10), scorer=1)
        with pytest.raises(ValueError):
            FilteringAnomalyModel(model=MovingAverageFilter(window=10), scorer="str")
        with pytest.raises(ValueError):
            FilteringAnomalyModel(
                model=MovingAverageFilter(window=10),
                scorer=MovingAverageFilter(window=10),
            )
        with pytest.raises(ValueError):
            FilteringAnomalyModel(
                model=MovingAverageFilter(window=10), scorer=[Norm(), "str"]
            )

    def test_univariate_ForecastingAnomalyModel(self):
        np.random.seed(40)

        np_train_slope = np.array(range(0, 100, 1))
        np_test_slope = np.array(range(0, 100, 1))

        np_test_slope[30:32] = 29
        np_test_slope[50:65] = np_test_slope[50:65] + 1
        np_test_slope[75:80] = np_test_slope[75:80] * 0.98

        train_series_slope = TimeSeries.from_values(np_train_slope)
        test_series_slope = TimeSeries.from_values(np_test_slope)

        np_anomalies = np.zeros(100)
        np_anomalies[30:32] = 1
        np_anomalies[50:55] = 1
        np_anomalies[70:80] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            test_series_slope.time_index, np_anomalies, columns=["is_anomaly"]
        )

        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=5),
            scorer=[
                Norm(),
                Difference(),
                WassersteinScorer(window_agg=False),
                KMeansScorer(k=5),
                KMeansScorer(window=10, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
                WassersteinScorer(window=15, window_agg=False),
            ],
        )

        anomaly_model.fit(train_series_slope, allow_model_training=True, start=0.1)
        score, pred_series = anomaly_model.score(
            test_series_slope, return_model_prediction=True, start=0.1
        )

        # check that NormScorer is the abs difference of pred_series and test_series_slope
        assert (
            pred_series - test_series_slope.slice_intersect(pred_series)
        ).__abs__() == Norm().score_from_prediction(test_series_slope, pred_series)

        # check that Difference is the difference of pred_series and test_series_slope
        assert test_series_slope.slice_intersect(
            pred_series
        ) - pred_series == Difference().score_from_prediction(
            test_series_slope, pred_series
        )

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, test_series_slope, metric="AUC_ROC", start=0.1
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, test_series_slope, metric="AUC_PR", start=0.1
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.773449920508744,
            0.40659777424483307,
            0.9153708133971291,
            0.7702702702702702,
            0.9135765550239234,
            0.7603338632750397,
            0.9153708133971292,
            0.9006591337099811,
        ]

        true_auc_pr = [
            0.4818991248542174,
            0.20023033665128342,
            0.9144135170539835,
            0.47953161438253644,
            0.9127969832903458,
            0.47039678636225957,
            0.9147124232933175,
            0.9604714100445533,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_univariate_FilteringAnomalyModel(self):
        np.random.seed(40)

        np_series_train = np.array(range(0, 100, 1)) + np.random.normal(
            loc=0, scale=1, size=100
        )
        np_series_test = np.array(range(0, 100, 1)) + np.random.normal(
            loc=0, scale=1, size=100
        )

        np_series_test[30:35] = np_series_test[30:35] + np.random.normal(
            loc=0, scale=10, size=5
        )
        np_series_test[50:60] = np_series_test[50:60] + np.random.normal(
            loc=0, scale=4, size=10
        )
        np_series_test[75:80] = np_series_test[75:80] + np.random.normal(
            loc=0, scale=3, size=5
        )

        train_series_noise = TimeSeries.from_values(np_series_train)
        test_series_noise = TimeSeries.from_values(np_series_test)

        np_anomalies = np.zeros(100)
        np_anomalies[30:35] = 1
        np_anomalies[50:60] = 1
        np_anomalies[75:80] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            test_series_noise.time_index, np_anomalies, columns=["is_anomaly"]
        )

        anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=5),
            scorer=[
                Norm(),
                Difference(),
                WassersteinScorer(window_agg=False),
                KMeansScorer(),
                KMeansScorer(window=10, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
                WassersteinScorer(window=15, window_agg=False),
            ],
        )
        anomaly_model.fit(train_series_noise)
        score, pred_series = anomaly_model.score(
            test_series_noise, return_model_prediction=True
        )

        # check that Difference is the difference of pred_series and test_series_noise
        assert test_series_noise.slice_intersect(
            pred_series
        ) - pred_series == Difference().score_from_prediction(
            test_series_noise, pred_series
        )

        # check that NormScorer is the abs difference of pred_series and test_series_noise
        assert (
            test_series_noise.slice_intersect(pred_series) - pred_series
        ).__abs__() == Norm().score_from_prediction(test_series_noise, pred_series)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, test_series_noise, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, test_series_noise, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.875625,
            0.5850000000000001,
            0.952127659574468,
            0.814375,
            0.9598646034816247,
            0.88125,
            0.9666344294003868,
            0.9731182795698925,
        ]

        true_auc_pr = [
            0.7691407907338141,
            0.5566414178265074,
            0.9720504927710986,
            0.741298584352156,
            0.9744855592642071,
            0.7808056518442923,
            0.9800621192517156,
            0.9911842778990486,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_univariate_covariate_ForecastingAnomalyModel(self):
        np.random.seed(40)

        day_week = [0, 1, 2, 3, 4, 5, 6]
        np_day_week = np.array(day_week * 10)

        np_train_series = 0.5 * np_day_week
        np_test_series = 0.5 * np_day_week

        np_test_series[30:35] = np_test_series[30:35] + np.random.normal(
            loc=0, scale=2, size=5
        )
        np_test_series[50:60] = np_test_series[50:60] + np.random.normal(
            loc=0, scale=1, size=10
        )

        covariates = TimeSeries.from_times_and_values(
            pd.date_range(start="1949-01-01", end="1949-03-11"), np_day_week
        )
        series_train = TimeSeries.from_times_and_values(
            pd.date_range(start="1949-01-01", end="1949-03-11"), np_train_series
        )
        series_test = TimeSeries.from_times_and_values(
            pd.date_range(start="1949-01-01", end="1949-03-11"), np_test_series
        )

        np_anomalies = np.zeros(70)
        np_anomalies[30:35] = 1
        np_anomalies[50:60] = 1
        ts_anomalies = TimeSeries.from_times_and_values(
            series_test.time_index, np_anomalies, columns=["is_anomaly"]
        )

        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=2, lags_future_covariates=[0]),
            scorer=[
                Norm(),
                Difference(),
                WassersteinScorer(window_agg=False),
                KMeansScorer(k=4),
                KMeansScorer(k=7, window=10, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
                WassersteinScorer(window=15, window_agg=False),
            ],
        )

        anomaly_model.fit(
            series_train,
            allow_model_training=True,
            future_covariates=covariates,
            start=0.2,
        )

        score, pred_series = anomaly_model.score(
            series_test,
            return_model_prediction=True,
            future_covariates=covariates,
            start=0.2,
        )

        # check that NormScorer is the abs difference of pred_series and series_test
        assert (
            series_test.slice_intersect(pred_series) - pred_series
        ).__abs__() == Norm().score_from_prediction(series_test, pred_series)

        # check that Difference is the difference of pred_series and series_test
        assert series_test.slice_intersect(
            pred_series
        ) - pred_series == Difference().score_from_prediction(series_test, pred_series)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, series_test, metric="AUC_ROC", start=0.2
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, series_test, metric="AUC_PR", start=0.2
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=score,
            window=[1, 1, 10, 1, 10, 1, 10, 15],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [1.0, 0.6, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]

        true_auc_pr = [
            1.0,
            0.6914399076961142,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.9999999999999999,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_multivariate_FilteringAnomalyModel(self):
        np.random.seed(40)

        data_1 = np.random.normal(0, 0.1, 100)
        data_2 = np.random.normal(0, 0.1, 100)

        mts_series_train = TimeSeries.from_values(
            np.dstack((data_1, data_2))[0], columns=["component 1", "component 2"]
        )

        data_1[15:20] = data_1[15:20] + np.random.normal(0, 0.9, 5)
        data_1[35:40] = data_1[35:40] + np.random.normal(0, 0.4, 5)

        data_2[50:55] = data_2[50:55] + np.random.normal(0, 0.7, 5)
        data_2[65:70] = data_2[65:70] + np.random.normal(0, 0.4, 5)

        data_1[80:85] = data_1[80:85] + np.random.normal(0, 0.6, 5)
        data_2[80:85] = data_2[80:85] + np.random.normal(0, 0.6, 5)

        data_1[93:98] = data_1[93:98] + np.random.normal(0, 0.6, 5)
        data_2[93:98] = data_2[93:98] + np.random.normal(0, 0.6, 5)
        mts_series_test = TimeSeries.from_values(
            np.dstack((data_1, data_2))[0], columns=["component 1", "component 2"]
        )

        np1_anomalies = np.zeros(len(data_1))
        np1_anomalies[15:20] = 1
        np1_anomalies[35:40] = 1
        np1_anomalies[80:85] = 1
        np1_anomalies[93:98] = 1

        np2_anomalies = np.zeros(len(data_2))
        np2_anomalies[50:55] = 1
        np2_anomalies[67:70] = 1
        np2_anomalies[80:85] = 1
        np2_anomalies[93:98] = 1

        np_anomalies = np.zeros(len(data_2))
        np_anomalies[15:20] = 1
        np_anomalies[35:40] = 1
        np_anomalies[50:55] = 1
        np_anomalies[67:70] = 1
        np_anomalies[80:85] = 1
        np_anomalies[93:98] = 1

        ts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index,
            np.dstack((np1_anomalies, np2_anomalies))[0],
            columns=["is_anomaly_1", "is_anomaly_2"],
        )

        mts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index, np_anomalies, columns=["is_anomaly"]
        )

        # first case: scorers that return univariate scores
        anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=10),
            scorer=[
                Norm(component_wise=False),
                WassersteinScorer(window_agg=False),
                WassersteinScorer(window=12, window_agg=False),
                KMeansScorer(),
                KMeansScorer(window=5, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=5, window_agg=False),
            ],
        )
        anomaly_model.fit(mts_series_train)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 12, 1, 5, 1, 5],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 12, 1, 5, 1, 5],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.8695436507936507,
            0.9737678855325913,
            0.9930555555555555,
            0.857638888888889,
            0.9639130434782609,
            0.8690476190476191,
            0.9630434782608696,
        ]

        true_auc_pr = [
            0.814256917602188,
            0.9945160041091712,
            0.9992086070916503,
            0.8054288542539664,
            0.9777504211642852,
            0.8164636240285442,
            0.9763049418985656,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

        # second case: scorers that return scorers that have the same width as the input
        anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=10),
            scorer=[
                Norm(component_wise=True),
                Difference(),
                WassersteinScorer(component_wise=True, window_agg=False),
                WassersteinScorer(window=12, component_wise=True, window_agg=False),
                KMeansScorer(component_wise=True),
                KMeansScorer(window=5, component_wise=True, window_agg=False),
                PyODScorer(model=KNN(), component_wise=True),
                PyODScorer(
                    model=KNN(), window=5, component_wise=True, window_agg=False
                ),
            ],
        )
        anomaly_model.fit(mts_series_train)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 12, 1, 5, 1, 5],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 12, 1, 5, 1, 5],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            [0.859375, 0.9200542005420054],
            [0.49875, 0.513550135501355],
            [0.997093023255814, 0.9536231884057971],
            [0.998960498960499, 0.9739795918367344],
            [0.8143750000000001, 0.8218157181571816],
            [0.9886148007590132, 0.94677734375],
            [0.830625, 0.9369918699186992],
            [0.9909867172675522, 0.94580078125],
        ]

        true_auc_pr = [
            [0.7213314465376244, 0.8191331553279771],
            [0.4172305056124696, 0.49249755343619195],
            [0.9975245098039216, 0.9741870252257915],
            [0.9992877492877493, 0.9865792868871687],
            [0.7095552075210219, 0.7591858780309868],
            [0.9827224901431558, 0.9402925739221939],
            [0.7095275592303261, 0.8313668186652059],
            [0.9858096294704315, 0.9391783485106905],
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_multivariate_ForecastingAnomalyModel(self):
        np.random.seed(40)

        data_sin = np.array([np.sin(x) for x in np.arange(0, 20 * np.pi, 0.2)])
        data_cos = np.array([np.cos(x) for x in np.arange(0, 20 * np.pi, 0.2)])

        mts_series_train = TimeSeries.from_values(
            np.dstack((data_sin, data_cos))[0], columns=["component 1", "component 2"]
        )

        data_sin[10:20] = 0
        data_cos[60:80] = 0

        data_sin[100:110] = 1
        data_cos[150:155] = 1

        data_sin[200:240] = 0.9 * data_cos[200:240]
        data_cos[200:240] = 0.9 * data_sin[200:240]

        data_sin[275:295] = data_sin[275:295] + np.random.normal(0, 0.1, 20)
        data_cos[275:295] = data_cos[275:295] + np.random.normal(0, 0.1, 20)

        mts_series_test = TimeSeries.from_values(
            np.dstack((data_sin, data_cos))[0], columns=["component 1", "component 2"]
        )

        np1_anomalies = np.zeros(len(data_sin))
        np1_anomalies[10:20] = 1
        np1_anomalies[100:110] = 1
        np1_anomalies[200:240] = 1
        np1_anomalies[275:295] = 1

        np2_anomalies = np.zeros(len(data_cos))
        np2_anomalies[60:80] = 1
        np2_anomalies[150:155] = 1
        np2_anomalies[200:240] = 1
        np2_anomalies[275:295] = 1

        np_anomalies = np.zeros(len(data_cos))
        np_anomalies[10:20] = 1
        np_anomalies[60:80] = 1
        np_anomalies[100:110] = 1
        np_anomalies[150:155] = 1
        np_anomalies[200:240] = 1
        np_anomalies[275:295] = 1

        ts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index,
            np.dstack((np1_anomalies, np2_anomalies))[0],
            columns=["is_anomaly_1", "is_anomaly_2"],
        )

        mts_anomalies = TimeSeries.from_times_and_values(
            mts_series_train.time_index, np_anomalies, columns=["is_anomaly"]
        )

        # first case: scorers that return univariate scores
        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10),
            scorer=[
                Norm(component_wise=False),
                WassersteinScorer(window_agg=False),
                WassersteinScorer(window=20, window_agg=False),
                KMeansScorer(),
                KMeansScorer(window=20, window_agg=False),
                PyODScorer(model=KNN()),
                PyODScorer(model=KNN(), window=10, window_agg=False),
            ],
        )
        anomaly_model.fit(mts_series_train, allow_model_training=True, start=0.1)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True, start=0.1
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, start=0.1, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            mts_anomalies, mts_series_test, start=0.1, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 20, 1, 20, 1, 10],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[mts_anomalies] * 7,
            pred_scores=scores,
            window=[1, 10, 20, 1, 20, 1, 10],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            0.9252575884154831,
            0.9130158730158731,
            0.9291228070175439,
            0.9252575884154832,
            0.9211929824561403,
            0.9252575884154831,
            0.915873015873016,
        ]

        true_auc_pr = [
            0.8389462532437767,
            0.9151621069238896,
            0.9685249535885079,
            0.8389462532437765,
            0.9662153835545242,
            0.8389462532437764,
            0.9212725256428517,
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

        # second case: scorers that return scorers that have the same width as the input
        anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10),
            scorer=[
                Norm(component_wise=True),
                Difference(),
                WassersteinScorer(component_wise=True, window_agg=False),
                WassersteinScorer(window=20, component_wise=True, window_agg=False),
                KMeansScorer(component_wise=True),
                KMeansScorer(window=20, component_wise=True, window_agg=False),
                PyODScorer(model=KNN(), component_wise=True),
                PyODScorer(
                    model=KNN(), window=10, component_wise=True, window_agg=False
                ),
            ],
        )
        anomaly_model.fit(mts_series_train, allow_model_training=True, start=0.1)

        scores, pred_series = anomaly_model.score(
            mts_series_test, return_model_prediction=True, start=0.1
        )

        # pred_series must be multivariate (same width as input)
        assert pred_series.width == mts_series_test.width

        # scores must be of the same length as the number of scorers
        assert len(scores) == len(anomaly_model.scorers)

        dict_auc_roc = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, start=0.1, metric="AUC_ROC"
        )
        dict_auc_pr = anomaly_model.eval_metric(
            ts_anomalies, mts_series_test, start=0.1, metric="AUC_PR"
        )

        auc_roc_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 20, 1, 20, 1, 10],
            metric="AUC_ROC",
        )

        auc_pr_from_scores = eval_metric_from_scores(
            anomalies=[ts_anomalies] * 8,
            pred_scores=scores,
            window=[1, 1, 10, 20, 1, 20, 1, 10],
            metric="AUC_PR",
        )

        # function eval_accuracy_from_scores and eval_accuracy must return an input of same length
        assert len(auc_roc_from_scores) == len(dict_auc_roc)
        assert len(auc_pr_from_scores) == len(dict_auc_pr)

        # function eval_accuracy_from_scores and eval_accuracy must return the same values
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, list(dict_auc_roc.values()), decimal=1
        )
        np.testing.assert_array_almost_equal(
            auc_pr_from_scores, list(dict_auc_pr.values()), decimal=1
        )

        true_auc_roc = [
            [0.8803738317757009, 0.912267218445167],
            [0.48898531375166887, 0.5758202778598878],
            [0.8375999073323295, 0.9162283996994741],
            [0.7798128494807715, 0.8739249880554228],
            [0.8803738317757008, 0.912267218445167],
            [0.7787287458632889, 0.8633540372670807],
            [0.8803738317757009, 0.9122672184451671],
            [0.8348777945094406, 0.9137061285821616],
        ]

        true_auc_pr = [
            [0.7123114333965317, 0.7579757115620807],
            [0.4447973021706103, 0.596776950584551],
            [0.744325434474558, 0.8984960888744328],
            [0.7653561450296187, 0.9233662817550338],
            [0.7123114333965317, 0.7579757115620807],
            [0.7852553779986415, 0.9185701347601994],
            [0.7123114333965319, 0.7579757115620807],
            [0.757208451057927, 0.8967178983419622],
        ]

        # check value of results
        np.testing.assert_array_almost_equal(
            auc_roc_from_scores, true_auc_roc, decimal=1
        )
        np.testing.assert_array_almost_equal(auc_pr_from_scores, true_auc_pr, decimal=1)

    def test_visualization(self):
        # test function show_anomalies() and show_anomalies_from_scores()
        forecasting_anomaly_model = ForecastingAnomalyModel(
            model=RegressionModel(lags=10), scorer=Norm()
        )
        forecasting_anomaly_model.fit(self.train, allow_model_training=True)

        filtering_anomaly_model = FilteringAnomalyModel(
            model=MovingAverageFilter(window=10), scorer=Norm()
        )

        self.show_anomalies_function(
            visualization_function=forecasting_anomaly_model.show_anomalies
        )
        self.show_anomalies_function(
            visualization_function=filtering_anomaly_model.show_anomalies
        )
        self.show_anomalies_function(visualization_function=show_anomalies_from_scores)

    def show_anomalies_function(self, visualization_function):
        # must input only one series
        with pytest.raises(ValueError) as err:
            visualization_function(series=[self.train, self.train])
        assert (
            str(err.value)
            == "`series` must be single `TimeSeries` or a sequence of `TimeSeries` of length `1`."
        )
        # input must be a series
        with pytest.raises(ValueError):
            visualization_function(series=[1, 2, 4])

        if visualization_function != show_anomalies_from_scores:
            # metric must be "AUC_ROC" or "AUC_PR"
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.anomalies,
                    metric="str",
                )
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.anomalies,
                    metric="auc_roc",
                )
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train, anomalies=self.anomalies, metric=1
                )

            # anomalies must be not none if metric is given
            with pytest.raises(ValueError):
                visualization_function(series=self.train, metric="AUC_ROC")

            # anomalies must be binary
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.test,
                    metric="AUC_ROC",
                )

            # anomalies must contain at least 1 anomaly if metric is given
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.only_0_anomalies,
                    metric="AUC_ROC",
                )

            # anomalies must contain at least 1 non-anomoulous timestamp
            # if metric is given
            with pytest.raises(ValueError):
                visualization_function(
                    series=self.train,
                    anomalies=self.only_1_anomalies,
                    metric="AUC_ROC",
                )
        else:
            # window must be a positive int
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train, pred_scores=self.test, window=-1
                )
            # window must smaller than the score series
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train, pred_scores=self.test, window=200
                )
            # must have the same nbr of windows than scores
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train, pred_scores=self.test, window=[1, 2]
                )
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train,
                    pred_scores=[self.test, self.test],
                    window=[1, 2, 1],
                )
            # nbr of names_of_scorers must match the nbr of scores
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train,
                    pred_scores=self.test,
                    names_of_scorers=["scorer1", "scorer2"],
                )
            with pytest.raises(ValueError):
                show_anomalies_from_scores(
                    series=self.train,
                    pred_scores=[self.test, self.test],
                    names_of_scorers=["scorer1", "scorer2", "scorer3"],
                )



import itertools

import pandas as pd
import pytest

import darts.utils.historical_forecasts.utils as hfc_utils
from darts.models import LinearRegressionModel
from darts.utils.timeseries_generation import linear_timeseries


class TestHistoricalForecastsUtils:
    model = LinearRegressionModel(lags=1)

    def test_historical_forecasts_check_kwargs(self):
        # `hfc_args` not part of `dict_kwargs` works
        hfc_args = {"a", "b"}
        dict_kwargs = {"c": 0, "d": 0}
        out = hfc_utils._historical_forecasts_check_kwargs(
            hfc_args=hfc_args,
            name_kwargs="some_name",
            dict_kwargs=dict_kwargs,
        )
        assert out == dict_kwargs

        # `hfc_args` is part of `dict_kwargs` fails
        with pytest.raises(ValueError):
            _ = hfc_utils._historical_forecasts_check_kwargs(
                hfc_args={"c"},
                name_kwargs="some_name",
                dict_kwargs=dict_kwargs,
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],  # retrain
            [True, False],  # show warnings
            [{}, {"some_fit_param": 0}],  # fit kwargs
            [{}, {"some_predict_param": 0}],  # predict kwargs
        ),
    )
    def test_historical_forecasts_sanitize_kwargs(self, config):
        retrain, show_warnings, fit_kwargs, pred_kwargs = config
        fit_kwargs_out, pred_kwargs_out = (
            hfc_utils._historical_forecasts_sanitize_kwargs(
                self.model,
                fit_kwargs=fit_kwargs,
                predict_kwargs=pred_kwargs,
                retrain=retrain,
                show_warnings=show_warnings,
            )
        )
        assert fit_kwargs_out == fit_kwargs
        assert pred_kwargs_out == pred_kwargs

    @pytest.mark.parametrize(
        "kwargs",
        [
            {
                "fit_kwargs": {"series": 0},
                "predict_kwargs": None,
                "retrain": True,
                "show_warnings": False,
            },
            {
                "fit_kwargs": None,
                "predict_kwargs": {"series": 0},
                "retrain": True,
                "show_warnings": False,
            },
        ],
    )
    def test_historical_forecasts_sanitize_kwargs_invalid(self, kwargs):
        with pytest.raises(ValueError):
            _ = hfc_utils._historical_forecasts_sanitize_kwargs(self.model, **kwargs)

    def test_historical_forecasts_check_start(self):
        """"""
        series = linear_timeseries(start=0, length=1)
        kwargs = {
            "start": 0,
            "start_format": "value",
            "series_start": 0,
            "ref_start": 0,
            "ref_end": 0,
            "stride": 0,
            "series_idx": 0,
            "is_historical_forecast": False,
        }
        # low enough start idx works with any kwargs
        hfc_utils._check_start(series, start_idx=0, **kwargs)

        # start idx >= len(series) raises error
        with pytest.raises(ValueError):
            hfc_utils._check_start(series, start_idx=1, **kwargs)

    @pytest.mark.parametrize(
        "config",
        [
            (True, pd.Timestamp("2000-01-01"), "value"),
            (True, 0.9, "value"),
            (True, 0.9, "position"),
            (True, 0, "position"),
            (True, 0, "value"),
            (True, -1, "position"),
            (False, pd.Timestamp("2000-01-01"), "value"),
            (False, 0.9, "value"),
            (False, 0.9, "position"),
            (False, 0, "position"),
            (False, -1, "position"),
        ],
    )
    def test_historical_forecasts_check_start_invalid(self, config):
        """"""
        is_dt, start, start_format = config
        series = linear_timeseries(start="2000-01-01" if is_dt else 0, length=1)
        series_start = series.start_time()
        kwargs = {
            "start": start,
            "start_format": start_format,
            "series_start": series_start,
            "ref_start": 0,
            "ref_end": 0,
            "stride": 0,
            "series_idx": 0,
            "is_historical_forecast": False,
        }

        # low enough start idx works with any kwargs
        with pytest.raises(ValueError) as err:
            hfc_utils._check_start(series, start_idx=1, **kwargs)

        # make sure we reach the expected error message and message is specific to input
        position_msg = f"position `{start}` corresponding to time "
        if start_format == "position" or is_dt and not isinstance(start, pd.Timestamp):
            assert position_msg in str(err.value)
        else:
            assert position_msg not in str(err.value)

    @pytest.mark.parametrize(
        "config",
        [
            (0, 0, 0),
            (1, 1, 1),
            (1, 10, 1),
            (-1, 1, 0),
            (-3, 1, 0),
            (-1, 2, 1),
            (-2, 2, 0),
            (-3, 2, 1),
        ],
    )
    def test_adjust_start(self, config):
        """Check relative start position adjustment."""
        start_rel, stride, start_expected = config
        assert hfc_utils._adjust_start(start_rel, stride) == start_expected
